# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import utils

class Entity:
    def __init__(self, line):
        # Each entity defines a TLV item in a QMI request or response.  The
        # entity's 'struct' field maps to the ID of a struct in Struct.txt,
        # which describes the fields of that struct.  For example:
        #
        # 33^"32,16"^"WDS/Start Network Interface Request/Primary DNS"^50021^-1
        #
        # The first field (33) indicates that this entity is valid for the QMI
        # message eDB2_ET_QMI_WDS_REQ. The "32,16" is a tuple
        # (eQMI_WDS_START_NET, 16), the first item of which (32) indicates the
        # general QMI service the entity is associated with (WDS, CTL, NAS, etc)
        # and the second item indicates the specific entity number (ie, the 'T'
        # in TLV).
        #
        # The 'struct' field (50021) points to:
        #
        # Struct.txt:50021^0^0^50118^""^-1^1^"4"
        #
        # which indicates that the only member of this struct is field #50118:
        #
        # Field.txt:50118^"IP V4 Address"^8^0^2^0
        #
        # which should be pretty self-explanatory.  Note that different entities
        # can point to the same struct.


        parts = line.split('^')
        if len(parts) < 4:
            raise Exception("Invalid entity line '%s'" % line)
        self.uniqueid = parts[0] + '.' + parts[1].replace('"', '').replace(',', '.')
        self.type = int(parts[0])  # eDB2EntityType

        self.key = parts[1].replace('"','')  # tuple of (eQMIMessageXXX, entity number)
        self.cmdno = int(self.key.split(",")[0])
        self.tlvno = int(self.key.split(",")[1])
        self.name = parts[2].replace('"', '')
        self.struct = int(parts[3])
        self.format = None
        self.internal = True
        self.extformat = None
        if len(parts) > 4:
            self.format = int(parts[4])
        if len(parts) > 5:
            self.internal = int(parts[5]) != 0
        if len(parts) > 6:
            self.extformat = int(parts[6])

    def validate(self, structs):
        if not structs.has_child(self.struct):
            raise Exception("Entity missing struct: %d" % self.struct)

    def emit(self, fields, structs, enums):
        if self.tlvno == 2 and self.name.find("/Result Code") > 0 and self.struct == 50000:
            # ignore this entity if it's a standard QMI result code struct
            return self.struct

        # Tell the struct this value is for to emit itself
        s = structs.get_child(self.struct)
        s.emit_header(self.name, self.cmdno, self.tlvno)
        s.emit(self.name, 0, 0, fields, structs, enums)
        return self.struct


class Entities(utils.DbFile):
    def __init__(self, path):
        self.byid = {}
        f = file(path + "Entity.txt")
        for line in f:
            ent = Entity(line)
            self.byid[ent.uniqueid] = ent

    def validate(self, structs):
        for e in self.byid.values():
            e.validate(structs)

    def emit(self, fields, structs, enums):
        # emit the standard status TLV struct
        print "struct qmi_result_code { /* QMI Result Code TLV (0x0002) */"
        print "\tgobi_qmi_results qmi_result; /* QMI Result */"
        print "\tgobi_qmi_errors qmi_error; /* QMI Error */"
        print "};"
        print ""

        structs_used = []
        for e in self.byid.values():
            sused = e.emit(fields, structs, enums)
            structs_used.append(sused)
        return structs_used




# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string

import utils
import VariableFactory
import TypeFactory

"""
The Field class takes care of handling Input and Output TLVs
"""
class Field:

    """
    Constructor
    """
    def __init__(self, service, prefix, dictionary, common_objects_dictionary, container_type, static):
        # The current QMI service
        self.service = service
        # The field prefix, usually the name of the Container,
        #  e.g. "Qmi Message Ctl Something Output"
        self.prefix = prefix
        # The name of the specific field, e.g. "Result"
        self.name = dictionary['name']
        # The specific TLV ID
        self.id = dictionary['id']
        # Overridden mandatory field, the default is calculated from id
        self._mandatory = dictionary.get('mandatory')
        # The type, which must always be "TLV"
        self.type = dictionary['type']
        # The container type, which must be either "Input" or "Output"
        self.container_type = container_type
        # Whether the whole field is internally used only
        self.static = static

        # Create the composed full name (prefix + name),
        #  e.g. "Qmi Message Ctl Something Output Result"
        self.fullname = dictionary['fullname'] if 'fullname' in dictionary else self.prefix + ' ' + self.name

        # libqmi version where the message was introduced
        self.since = dictionary['since'] if 'since' in dictionary else None
        if self.since is None:
            raise ValueError('TLV ' + self.fullname + ' requires a "since" tag specifying the major version where it was introduced')

        # TLVs can no longer be structs, they must be sequences instead
        if dictionary['format'] == 'struct':
            raise ValueError('TLV ' + self.fullname + ' cannot have type "struct": use "sequence" instead')

        # Create our variable object
        self.variable = VariableFactory.create_variable(self.service, dictionary, self.fullname, self.container_type)

        # Create the variable name within the Container
        self.variable_name = 'arg_' + utils.build_underscore_name(self.name).lower()

        # Create the ID enumeration name
        self.id_enum_name = utils.build_underscore_name(self.prefix + ' TLV ' + self.name).upper()

        # Output Fields may have prerequisites
        self.prerequisites = []
        if 'prerequisites' in dictionary:
            self.prerequisites = dictionary['prerequisites']
            # First, look for references to common types
            for prerequisite_dictionary in self.prerequisites:
                if 'common-ref' in prerequisite_dictionary:
                    for common in common_objects_dictionary:
                        if common['type'] == 'prerequisite' and \
                           common['common-ref'] == prerequisite_dictionary['common-ref']:
                           # Replace the reference with a copy of the common dictionary
                           copy = dict(common)
                           self.prerequisites.remove(prerequisite_dictionary)
                           self.prerequisites.append(copy)
                           break
                    else:
                        raise RuntimeError('Common type \'%s\' not found' % prerequisite_dictionary['name'])

        if 'personal-info' in dictionary:
            self.personal_info = True;
        else:
            self.personal_info = False;


    @property
    def mandatory(self):
        if self._mandatory is None:
            return int(self.id, 0) < 0x10
        return self._mandatory == 'yes'


    """
    Emit new types required by this field
    """
    def emit_types(self, hfile, cfile):
        if not TypeFactory.is_type_emitted(self.fullname):
            TypeFactory.set_type_emitted(self.fullname)
            self.variable.emit_types(hfile, cfile, self.since, False)
            if self.variable.needs_compat_gir and self.service != 'CTL':
                self.variable.emit_types_gir(hfile, cfile, self.since)


    """
    Common getter logic
    """
    def emit_getter_common(self, hfile, cfile, dec, doc, imp, since, suffix_gir):
        translations = { 'name'                : self.name,
                         'variable_name'       : self.variable_name,
                         'variable_getter_dec' : dec,
                         'variable_getter_doc' : doc,
                         'variable_getter_imp' : imp,
                         'underscore'          : utils.build_underscore_name(self.name),
                         'suffix_gir'          : suffix_gir,
                         'prefix_camelcase'    : utils.build_camelcase_name(self.prefix),
                         'prefix_underscore'   : utils.build_underscore_name(self.prefix),
                         'since'               : since,
                         'static'              : 'static ' if self.static else '' }

        # Emit the getter header
        template = '\n'
        if not self.static and self.service != 'CTL':
            template += (
                '\n'
                '/**\n')

            if self.variable.needs_compat_gir:
                # Rename the GIR specific method to the original method name, and
                # skip the original method completely
                if suffix_gir == '':
                    template += (
                        ' * ${prefix_underscore}_get_${underscore}: (skip)\n')
                else:
                    template += (
                        ' * ${prefix_underscore}_get_${underscore}${suffix_gir}: (rename-to ${prefix_underscore}_get_${underscore})\n')
            else:
                if suffix_gir != '':
                    raise RuntimeError('GIR suffix given and variable does not need compat GIR support')
                template += (
                    ' * ${prefix_underscore}_get_${underscore}:\n')

            template += (
                ' * @self: a #${prefix_camelcase}.\n'
                '${variable_getter_doc}'
                ' * @error: Return location for error or %NULL.\n'
                ' *\n'
                ' * Get the \'${name}\' field from @self.\n')

            if self.variable.needs_compat_gir:
                if suffix_gir != '':
                    template += (
                        ' *\n'
                        ' * Version of ${prefix_underscore}_get_${underscore}() using arrays of pointers to\n'
                        ' * structs instead of arrays of structs, for easier binding in other languages.\n')

            template += (
                ' *\n'
                ' * Returns: (skip): %TRUE if the field is found, %FALSE otherwise.\n'
                ' *\n'
                ' * Since: ${since}\n'
                ' */\n')
        template += (
            '${static}gboolean ${prefix_underscore}_get_${underscore}${suffix_gir} (\n'
            '    ${prefix_camelcase} *self,\n'
            '${variable_getter_dec}'
            '    GError **error);\n')
        hfile.write(string.Template(template).substitute(translations))

        # Emit the getter source
        template = (
            '\n'
            '${static}gboolean\n'
            '${prefix_underscore}_get_${underscore}${suffix_gir} (\n'
            '    ${prefix_camelcase} *self,\n'
            '${variable_getter_dec}'
            '    GError **error)\n'
            '{\n'
            '    g_return_val_if_fail (self != NULL, FALSE);\n'
            '\n'
            '    if (!self->${variable_name}_set) {\n'
            '        g_set_error (error,\n'
            '                     QMI_CORE_ERROR,\n'
            '                     QMI_CORE_ERROR_TLV_NOT_FOUND,\n'
            '                     "Field \'${name}\' was not found in the message");\n'
            '        return FALSE;\n'
            '    }\n'
            '\n'
            '${variable_getter_imp}'
            '\n'
            '    return TRUE;\n'
            '}\n')
        cfile.write(string.Template(template).substitute(translations))


    """
    Emit the method responsible for declaring variable(s) for this TLV in the
    input/output container
    """
    def emit_variable_declaration(self, cfile):

        template = self.variable.build_variable_declaration('    ', self.variable_name)
        cfile.write(template)

        if self.variable.needs_compat_gir and self.service != 'CTL':
            template = self.variable.build_variable_declaration_gir('    ', self.variable_name)
            cfile.write(template)


    """
    Emit the method responsible for getting this TLV from the input/output
    container
    """
    def emit_getter(self, hfile, cfile):
        input_variable_name = 'value_' + utils.build_underscore_name(self.name)
        dec = self.variable.build_getter_declaration('    ', input_variable_name)
        doc = self.variable.build_getter_documentation(' * ', input_variable_name)
        imp = self.variable.build_getter_implementation('    ', 'self->' + self.variable_name, input_variable_name)
        since = self.since
        self.emit_getter_common(hfile, cfile, dec, doc, imp, since, '')

        if self.variable.needs_compat_gir and self.service != 'CTL':
            input_variable_name = 'value_' + utils.build_underscore_name(self.name)
            dec = self.variable.build_getter_declaration_gir('    ', input_variable_name)
            doc = self.variable.build_getter_documentation_gir(' * ', input_variable_name)
            imp = self.variable.build_getter_implementation_gir('    ', 'self->' + self.variable_name, input_variable_name)
            since = self.since if utils.version_compare('1.32',self.since) > 0 else '1.32'
            self.emit_getter_common(hfile, cfile, dec, doc, imp, since, '_gir')


    """
    Common setter logic
    """
    def emit_setter_common(self, hfile, cfile, dec, doc, imp, since, suffix_gir):
        translations = { 'name'                : self.name,
                         'variable_name'       : self.variable_name,
                         'variable_setter_dec' : dec,
                         'variable_setter_doc' : doc,
                         'variable_setter_imp' : imp,
                         'underscore'          : utils.build_underscore_name(self.name),
                         'prefix_camelcase'    : utils.build_camelcase_name(self.prefix),
                         'prefix_underscore'   : utils.build_underscore_name(self.prefix),
                         'since'               : since,
                         'suffix_gir'          : suffix_gir,
                         'static'              : 'static ' if self.static else '' }

        # Emit the setter header
        template = '\n'
        if not self.static and self.service != 'CTL':
            template += (
                '\n'
                '/**\n')

            if self.variable.needs_compat_gir:
                # Rename the GIR specific method to the original method name, and
                # skip the original method completely
                if suffix_gir == '':
                    template += (
                        ' * ${prefix_underscore}_set_${underscore}: (skip)\n')
                else:
                    template += (
                        ' * ${prefix_underscore}_set_${underscore}${suffix_gir}: (rename-to ${prefix_underscore}_set_${underscore})\n')
            else:
                if suffix_gir != '':
                    raise RuntimeError('GIR suffix given and variable does not need compat GIR support')
                template += (
                    ' * ${prefix_underscore}_set_${underscore}:\n')

            template += (
                ' * @self: a #${prefix_camelcase}.\n'
                '${variable_setter_doc}'
                ' * @error: Return location for error or %NULL.\n'
                ' *\n'
                ' * Set the \'${name}\' field in the message.\n'
                ' *\n'
                ' * Returns: (skip): %TRUE if @value was successfully set, %FALSE otherwise.\n'
                ' *\n'
                ' * Since: ${since}\n'
                ' */\n')
        template += (
            '${static}gboolean ${prefix_underscore}_set_${underscore}${suffix_gir} (\n'
            '    ${prefix_camelcase} *self,\n'
            '${variable_setter_dec}'
            '    GError **error);\n')
        hfile.write(string.Template(template).substitute(translations))

        # Emit the setter source
        template = (
            '\n'
            '${static}gboolean\n'
            '${prefix_underscore}_set_${underscore}${suffix_gir} (\n'
            '    ${prefix_camelcase} *self,\n'
            '${variable_setter_dec}'
            '    GError **error)\n'
            '{\n'
            '    g_return_val_if_fail (self != NULL, FALSE);\n'
            '\n'
            '${variable_setter_imp}'
            '    self->${variable_name}_set = TRUE;\n'
            '\n'
            '    return TRUE;\n'
            '}\n')
        cfile.write(string.Template(template).substitute(translations))

    """
    Emit the method responsible for setting this TLV in the input/output
    container
    """
    def emit_setter(self, hfile, cfile):
        input_variable_name = 'value_' + utils.build_underscore_name(self.name)
        dec = self.variable.build_setter_declaration('    ', input_variable_name)
        doc = self.variable.build_setter_documentation(' * ', input_variable_name)
        imp = self.variable.build_setter_implementation('    ', input_variable_name, 'self->' + self.variable_name)
        since = self.since
        self.emit_setter_common(hfile, cfile, dec, doc, imp, since, '')

        if self.variable.needs_compat_gir and self.service != 'CTL':
            input_variable_name = 'value_' + utils.build_underscore_name(self.name)
            dec = self.variable.build_setter_declaration_gir('    ', input_variable_name)
            doc = self.variable.build_setter_documentation_gir(' * ', input_variable_name)
            imp = self.variable.build_setter_implementation_gir('    ', input_variable_name, 'self->' + self.variable_name)
            since = self.since if utils.version_compare('1.32',self.since) > 0 else '1.32'
            self.emit_setter_common(hfile, cfile, dec, doc, imp, since, '_gir')


    """
    Emit the code responsible for adding the TLV to the QMI message
    """
    def emit_input_tlv_add(self, f, line_prefix):
        translations = { 'name'          : self.name,
                         'tlv_id'        : self.id_enum_name,
                         'variable_name' : self.variable_name,
                         'lp'            : line_prefix }

        template = (
            '${lp}gsize tlv_offset;\n'
            '\n'
            '${lp}if (!(tlv_offset = qmi_message_tlv_write_init (self, (guint8)${tlv_id}, error))) {\n'
            '${lp}    g_prefix_error (error, "Cannot initialize TLV \'${name}\': ");\n'
            '${lp}    return NULL;\n'
            '${lp}}\n'
            '\n')
        f.write(string.Template(template).substitute(translations))

        # Now, write the contents of the variable into the buffer
        self.variable.emit_buffer_write(f, line_prefix, self.name, 'input->' + self.variable_name)

        template = (
            '\n'
            '${lp}if (!qmi_message_tlv_write_complete (self, tlv_offset, error)) {\n'
            '${lp}    g_prefix_error (error, "Cannot complete TLV \'${name}\': ");\n'
            '${lp}    return NULL;\n'
            '${lp}}\n')
        f.write(string.Template(template).substitute(translations))


    """
    Emit the code responsible for checking prerequisites in output TLVs
    """
    def emit_output_prerequisite_check(self, f, line_prefix):
        if self.prerequisites == []:
            f.write('%s/* No Prerequisites for field */\n' % line_prefix)
            return

        for prerequisite in self.prerequisites:
            translations = { 'lp'                     : line_prefix,
                             'prerequisite_field'     : utils.build_underscore_name(prerequisite['field']),
                             'prerequisite_operation' : prerequisite['operation'],
                             'prerequisite_value'     : prerequisite['value'] }
            template = (
                '${lp}/* Prerequisite.... */\n'
                '${lp}if (!(self->arg_${prerequisite_field} ${prerequisite_operation} ${prerequisite_value}))\n'
                '${lp}    break;\n')
            f.write(string.Template(template).substitute(translations))


    """
    Emit the code responsible for retrieving the TLV from the QMI message
    """
    def emit_output_tlv_get(self, f, line_prefix):
        tlv_out = utils.build_underscore_name (self.fullname) + '_out'
        error = 'error' if self.mandatory else 'NULL'
        translations = { 'name'                 : self.name,
                         'container_underscore' : utils.build_underscore_name (self.prefix),
                         'tlv_out'              : tlv_out,
                         'tlv_id'               : self.id_enum_name,
                         'variable_name'        : self.variable_name,
                         'lp'                   : line_prefix,
                         'error'                : error }

        template = (
            '${lp}gsize offset = 0;\n'
            '${lp}gsize init_offset;\n'
            '\n'
            '${lp}if ((init_offset = qmi_message_tlv_read_init (message, ${tlv_id}, NULL, ${error})) == 0) {\n')

        if self.mandatory:
            template += (
                '${lp}    g_prefix_error (${error}, "Couldn\'t get the mandatory ${name} TLV: ");\n'
                '${lp}    ${container_underscore}_unref (self);\n'
                '${lp}    return NULL;\n')
        else:
            template += (
                '${lp}    goto ${tlv_out};\n')

        template += (
            '${lp}}\n')

        f.write(string.Template(template).substitute(translations))

        # Now, read the contents of the buffer into the variable
        self.variable.emit_buffer_read(f, line_prefix, tlv_out, error, 'self->' + self.variable_name)

        template = (
            '\n'
            '${lp}/* The remaining size of the buffer needs to be 0 if we successfully read the TLV */\n'
            '${lp}if ((offset = qmi_message_tlv_read_remaining_size (message, init_offset, offset)) > 0) {\n'
            '${lp}    g_warning ("Left \'%" G_GSIZE_FORMAT "\' bytes unread when getting the \'${name}\' TLV", offset);\n'
            '${lp}}\n'
            '\n'
            '${lp}self->${variable_name}_set = TRUE;\n'
            '\n'
            '${tlv_out}:\n')
        if self.mandatory:
            template += (
                '${lp}if (!self->${variable_name}_set) {\n'
                '${lp}    ${container_underscore}_unref (self);\n'
                '${lp}    return NULL;\n'
                '${lp}}\n')
        else:
            template += (
                '${lp};\n')
        f.write(string.Template(template).substitute(translations))


    """
    Emit the method responsible for creating a printable representation of the TLV
    """
    def emit_tlv_helpers(self, f):
        if TypeFactory.helpers_emitted(self.fullname):
            return

        TypeFactory.set_helpers_emitted(self.fullname)

        translations = { 'name'       : self.name,
                         'tlv_id'     : self.id_enum_name,
                         'underscore' : utils.build_underscore_name (self.fullname) }

        template = (
            '\n'
            'static gchar *\n'
            '${underscore}_get_printable (\n'
            '    QmiMessage *message,\n'
            '    const gchar *line_prefix)\n'
            '{\n'
            '    gsize offset = 0;\n'
            '    gsize init_offset;\n'
            '    GString *printable;\n'
            '    GError *error = NULL;\n'
            '\n'
            '    if ((init_offset = qmi_message_tlv_read_init (message, ${tlv_id}, NULL, NULL)) == 0)\n'
            '        return NULL;\n'
            '\n'
            '    printable = g_string_new ("");\n')
        f.write(string.Template(template).substitute(translations))

        # Now, read the contents of the buffer into the printable representation
        self.variable.emit_get_printable(f, '    ', self.personal_info)

        template = (
            '\n'
            '    if ((offset = qmi_message_tlv_read_remaining_size (message, init_offset, offset)) > 0)\n'
            '        g_string_append_printf (printable, "Additional unexpected \'%" G_GSIZE_FORMAT "\' bytes", offset);\n'
            '\n'
            'out:\n'
            '    if (error) {\n'
            '        g_string_append_printf (printable, " ERROR: %s", error->message);\n'
            '        g_error_free (error);\n'
            '    }\n'
            '    return g_string_free (printable, FALSE);\n'
            '}\n')
        f.write(string.Template(template).substitute(translations))


    """
    Add sections
    """
    def add_sections(self, sections):
        translations = { 'underscore'        : utils.build_underscore_name(self.name),
                         'prefix_camelcase'  : utils.build_camelcase_name(self.prefix),
                         'prefix_underscore' : utils.build_underscore_name(self.prefix) }

        if not TypeFactory.is_section_emitted(self.fullname):
            TypeFactory.set_section_emitted(self.fullname)
            self.variable.add_sections(sections)

        # Public methods
        template = (
            '${prefix_underscore}_get_${underscore}\n')
        if self.variable.needs_compat_gir and self.service != 'CTL':
            template += (
                '${prefix_underscore}_get_${underscore}_gir\n')
        if self.container_type == 'Input':
            template += (
                '${prefix_underscore}_set_${underscore}\n')
            if self.variable.needs_compat_gir and self.service != 'CTL':
                template += (
                    '${prefix_underscore}_set_${underscore}_gir\n')
        sections['public-methods'] += string.Template(template).substitute(translations)



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils
from Variable import Variable

"""
Variable type for Strings ('string' format)
"""
class VariableString(Variable):

    def __init__(self, service, dictionary):

        # Call the parent constructor
        Variable.__init__(self, service, dictionary)

        self.private_format = 'gchar *'
        self.public_format = self.private_format
        self.element_type = 'utf8'

        # Public format for the GIR array compat methods in 1.32
        self.private_format_gir = self.private_format
        self.public_format_gir = self.public_format
        self.element_type_gir = self.element_type

        if 'fixed-size' in dictionary:
            self.is_fixed_size = True
            # Fixed-size strings
            self.needs_dispose = False
            self.length_prefix_size = 0
            self.n_size_prefix_bytes = 0
            self.fixed_size = dictionary['fixed-size']
            self.max_size = ''
        else:
            self.is_fixed_size = False
            self.fixed_size = '-1'
            # Variable-length strings in heap
            self.needs_dispose = True
            self.clear_method = 'qmi_helpers_clear_string'
            self.free_method_gir = 'g_free'
            if 'size-prefix-format' in dictionary:
                if dictionary['size-prefix-format'] == 'guint8':
                    self.length_prefix_size = 8
                    self.n_size_prefix_bytes = 1
                elif dictionary['size-prefix-format'] == 'guint16':
                    self.length_prefix_size = 16
                    self.n_size_prefix_bytes = 2
                else:
                    raise ValueError('Invalid size prefix format (%s): not guint8 or guint16' % dictionary['size-prefix-format'])
            # Strings which are given as the full value of a TLV and which don't have
            # a explicit 'size-prefix-format' will NOT have a length prefix
            elif 'type' in dictionary and dictionary['type'] == 'TLV':
                self.length_prefix_size = 0
                self.n_size_prefix_bytes = 0
            else:
                # Default to UINT8
                self.length_prefix_size = 8
                self.n_size_prefix_bytes = 1
            self.max_size = dictionary['max-size'] if 'max-size' in dictionary else ''


    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        translations = { 'lp'            : line_prefix,
                         'tlv_out'       : tlv_out,
                         'variable_name' : variable_name,
                         'error'         : error }

        if self.is_fixed_size:
            translations['fixed_size'] = self.fixed_size

            # Fixed sized strings exposed in public fields will need to be
            # explicitly allocated in heap
            if self.public:
                translations['fixed_size_plus_one'] = int(self.fixed_size) + 1
                template = (
                    '${lp}${variable_name} = g_malloc (${fixed_size_plus_one});\n'
                    '${lp}if (!qmi_message_tlv_read_fixed_size_string (message, init_offset, &offset, ${fixed_size}, &${variable_name}[0], ${error})) {\n'
                    '${lp}    g_free (${variable_name});\n'
                    '${lp}    ${variable_name} = NULL;\n'
                    '${lp}    goto ${tlv_out};\n'
                    '${lp}}\n'
                    '${lp}${variable_name}[${fixed_size}] = \'\\0\';\n')
            else:
                template = (
                    '${lp}if (!qmi_message_tlv_read_fixed_size_string (message, init_offset, &offset, ${fixed_size}, &${variable_name}[0], ${error}))\n'
                    '${lp}    goto ${tlv_out};\n'
                    '${lp}${variable_name}[${fixed_size}] = \'\\0\';\n')
        else:
            translations['n_size_prefix_bytes'] = self.n_size_prefix_bytes
            translations['max_size'] = self.max_size if self.max_size != '' else '0'
            template = (
                '${lp}if (!qmi_message_tlv_read_string (message, init_offset, &offset, ${n_size_prefix_bytes}, ${max_size}, &(${variable_name}), ${error}))\n'
                '${lp}    goto ${tlv_out};\n')
        f.write(string.Template(template).substitute(translations))


    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        translations = { 'lp'                  : line_prefix,
                         'tlv_name'            : tlv_name,
                         'variable_name'       : variable_name,
                         'fixed_size'          : self.fixed_size,
                         'n_size_prefix_bytes' : self.n_size_prefix_bytes }

        template = (
            '${lp}if (!qmi_message_tlv_write_string (self, ${n_size_prefix_bytes}, ${variable_name}, ${fixed_size}, error)) {\n'
            '${lp}    g_prefix_error (error, "Cannot write string in TLV \'${tlv_name}\': ");\n'
            '${lp}    return NULL;\n'
            '${lp}}\n')

        f.write(string.Template(template).substitute(translations))


    def emit_get_printable(self, f, line_prefix, is_personal):
        translations = { 'lp' : line_prefix }

        if self.is_fixed_size:
            translations['fixed_size'] = self.fixed_size
            translations['fixed_size_plus_one'] = int(self.fixed_size) + 1
            template = (
                '\n'
                '${lp}{\n'
                '${lp}    gchar tmp[${fixed_size_plus_one}] = { \'\\0\' };\n'
                '\n'
                '${lp}    if (!qmi_message_tlv_read_fixed_size_string (message, init_offset, &offset, ${fixed_size}, &tmp[0], &error))\n'
                '${lp}        goto out;\n')
        else:
            translations['n_size_prefix_bytes'] = self.n_size_prefix_bytes
            translations['max_size'] = self.max_size if self.max_size != '' else '0'
            template = (
                '\n'
                '${lp}{\n'
                '${lp}    g_autofree gchar *tmp = NULL;\n'
                '\n'
                '${lp}    if (!qmi_message_tlv_read_string (message, init_offset, &offset, ${n_size_prefix_bytes}, ${max_size}, &tmp, &error))\n'
                '${lp}        goto out;\n')

        if self.personal_info or is_personal:
            translations['if_show_field'] = 'if (qmi_utils_get_show_personal_info ()) '
        else:
            translations['if_show_field'] = ''

        template += (
            '${lp}    ${if_show_field}{\n'
            '${lp}        g_string_append (printable, tmp);\n')

        if self.personal_info or is_personal:
            template += (
                '${lp}    } else {\n'
                '${lp}        g_string_append_printf (printable, "\'###\'");\n')

        template += (
            '${lp}    }\n'
            '${lp}}\n')

        f.write(string.Template(template).substitute(translations))


    def build_variable_declaration(self, line_prefix, variable_name):
        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        if self.is_fixed_size:
            translations['fixed_size_plus_one'] = int(self.fixed_size) + 1
            template = (
                '${lp}gchar ${name}[${fixed_size_plus_one}];\n')
        else:
            template = (
                '${lp}gchar *${name};\n')
        return string.Template(template).substitute(translations)


    def build_struct_field_declaration(self, line_prefix, variable_name):
        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        # Fixed sized strings given in public structs are given as pointers,
        # instead of as fixed-sized arrays directly in the struct.
        template = (
            '${lp}gchar *${name};\n')
        return string.Template(template).substitute(translations)


    def build_getter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = (
            '${lp}const gchar **${name},\n')
        return string.Template(template).substitute(translations)


    def build_getter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = (
            '${lp}@${name}: (out)(optional)(transfer none): a placeholder for the output constant string, or %NULL if not required.\n')
        return string.Template(template).substitute(translations)


    def build_getter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        template = (
            '${lp}if (${to})\n'
            '${lp}    *${to} = ${from};\n')
        return string.Template(template).substitute(translations)


    def build_setter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = (
            '${lp}const gchar *${name},\n')
        return string.Template(template).substitute(translations)


    def build_setter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        if self.is_fixed_size:
            translations['fixed_size'] = self.fixed_size
            template = (
                '${lp}@${name}: a constant string of exactly ${fixed_size} characters.\n')
        elif self.max_size != '':
            translations['max_size'] = self.max_size
            template = (
                '${lp}@${name}: a constant string with a maximum length of ${max_size} characters.\n')
        else:
            template = (
                '${lp}@${name}: a constant string.\n')
        return string.Template(template).substitute(translations)


    def build_setter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        if self.is_fixed_size:
            translations['fixed_size'] = self.fixed_size
            template = (
                '${lp}if (!${from} || strlen (${from}) != ${fixed_size}) {\n'
                '${lp}    g_set_error (error,\n'
                '${lp}                 QMI_CORE_ERROR,\n'
                '${lp}                 QMI_CORE_ERROR_INVALID_ARGS,\n'
                '${lp}                 "Input variable \'${from}\' must be ${fixed_size} characters long");\n'
                '${lp}    return FALSE;\n'
                '${lp}}\n'
                '${lp}memcpy (${to}, ${from}, ${fixed_size});\n'
                '${lp}${to}[${fixed_size}] = \'\\0\';\n')
        else:
            template = ''
            if self.max_size != '':
                translations['max_size'] = self.max_size
                template += (
                    '${lp}if (${from} && strlen (${from}) > ${max_size}) {\n'
                    '${lp}    g_set_error (error,\n'
                    '${lp}                 QMI_CORE_ERROR,\n'
                    '${lp}                 QMI_CORE_ERROR_INVALID_ARGS,\n'
                    '${lp}                 "Input variable \'${from}\' must be less than ${max_size} characters long");\n'
                    '${lp}    return FALSE;\n'
                    '${lp}}\n')
            template += (
                '${lp}g_free (${to});\n'
                '${lp}${to} = g_strdup (${from} ? ${from} : "");\n')

        return string.Template(template).substitute(translations)


    def build_struct_field_documentation(self, line_prefix, variable_name):
        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        if self.is_fixed_size:
            translations['fixed_size'] = self.fixed_size
            template = (
                '${lp}@${name}: a string of exactly ${fixed_size} characters.\n')
        elif self.max_size != '':
            translations['max_size'] = self.max_size
            template = (
                '${lp}@${name}: a string with a maximum length of ${max_size} characters.\n')
        else:
            template = (
                '${lp}@${name}: a string.\n')
        return string.Template(template).substitute(translations)


    def build_dispose(self, line_prefix, variable_name):
        # Fixed-size strings don't need dispose
        if self.is_fixed_size and not self.public:
            return ''

        translations = { 'lp'            : line_prefix,
                         'variable_name' : variable_name }

        template = (
            '${lp}g_clear_pointer (&${variable_name}, (GDestroyNotify)g_free);\n')
        return string.Template(template).substitute(translations)


    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'                 : line_prefix,
                         'variable_name_from' : variable_name_from,
                         'variable_name_to'   : variable_name_to }

        template = (
            '${lp}${variable_name_to} = g_strdup (${variable_name_from});\n')
        return string.Template(template).substitute(translations)


    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_copy_gir (line_prefix, variable_name_from, variable_name_to)


    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_copy_gir (line_prefix, variable_name_from, variable_name_to)


    def flag_public(self):
        # Call the parent method
        Variable.flag_public(self)
        # Fixed-sized strings will need dispose if they are in the public header
        if self.is_fixed_size:
            self.needs_dispose = True



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils

"""
Base class for every variable type defined in the database
"""
class Variable:

    """
    Constructor with common variable handling
    """
    def __init__(self, service, dictionary):
        """
        The current QMI service
        """
        self.service = service

        """
        Variables can define specific public and private formats to be used.
        The public format will be that used in the generated interface file,
        while the private one will only be used internally.
        """
        self.format = dictionary['format']
        self.public_format = None
        self.private_format = None

        """
        Element type to be used in introspection annotations.
        """
        self.element_type = None

        """
        Whether the variable is visible in public API or is reserved
        """
        self.visible = False if ('visible' in dictionary and dictionary['visible'] == 'no') else True

        """
        Variables that get allocated in heap need to get properly disposed.
        """
        self.needs_dispose = False

        """
        Variables that get allocated in heap need to have a clear method so that it
        can be used as part of an array of this variable type.
        """
        self.clear_method = ''

        """
        Custom endianness configuration for a specific variable; if none given, defaults
        to host endian.
        """
        self.endian = "QMI_ENDIAN_LITTLE"
        if 'endian' in dictionary:
            endian = dictionary['endian']
            if endian == 'network' or endian == 'big':
                self.endian = "QMI_ENDIAN_BIG"
            elif endian == 'little':
                pass
            else:
                raise ValueError("Invalid endian value %s" % endian)

        """
        Initially all variables are flagged as not being public
        """
        self.public = False

        # -----------------------------------------------------------------------------
        # Variables to support the new GIR array/struct compat methods in 1.32

        """
        Flag specifying whether a given variable contains any kind of array, indicating
        the GIR array/struct compat methods in 1.32 are required. There is no need to
        flag structs independently, because it is ensured that structs are always array
        members.
        """
        self.needs_compat_gir = False

        """
        Formats and element types
        """
        self.public_format_gir = None
        self.private_format_gir = None
        self.element_type_gir = None

        """
        Types involved in GIR support may need additional free() and new() methods.
        """
        self.new_method_gir = ''
        self.free_method_gir = ''

        """
        Whether the variable is personal info or not
        """
        if 'personal-info' in dictionary:
            self.personal_info = True;
        else:
            self.personal_info = False;

        """
        Whether the variable contains personal info or not
        """
        self.contains_personal_info = self.personal_info

    """
    Emits the code to declare specific new types required by the variable.
    """
    def emit_types(self, hfile, cfile, since, static):
        pass

    """
    Emits the code involved in reading the variable from the raw byte stream
    into the specific private format.
    """
    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        pass

    """
    Emits the code involved in writing the variable to the raw byte stream
    from the specific private format.
    """
    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        pass

    """
    Emits the code to get the contents of the given variable as a printable string.
    """
    def emit_get_printable(self, f, line_prefix, is_personal):
        pass

    """
    Builds the code to include the declaration of a variable of this kind,
    used when generating input/output bundles.
    """
    def build_variable_declaration(self, line_prefix, variable_name):
        return ''

    """
    Builds the code to include in the getter method declaration for this kind of variable.
    """
    def build_getter_declaration(self, line_prefix, variable_name):
        return ''

    """
    Builds the documentation of the getter code
    """
    def build_getter_documentation(self, line_prefix, variable_name):
        return ''

    """
    Builds the code to implement getting this kind of variable.
    """
    def build_getter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        return ''

    """
    Builds the code to include in the setter method for this kind of variable.
    """
    def build_setter_declaration(self, line_prefix, variable_name):
        return ''

    """
    Builds the documentation of the setter code
    """
    def build_setter_documentation(self, line_prefix, variable_name):
        return ''

    """
    Builds the code to implement setting this kind of variable.
    """
    def build_setter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        return ''

    """
    Builds the code to include the declaration of a variable of this kind
    as a field in a public struct
    """
    def build_struct_field_declaration(self, line_prefix, variable_name):
        return ''

    """
    Documentation for the struct field
    """
    def build_struct_field_documentation(self, line_prefix, variable_name):
        return ''

    """
    Emits the code to dispose the variable.
    """
    def build_dispose(self, line_prefix, variable_name):
        return ''

    """
    Add sections
    """
    def add_sections(self, sections):
        pass

    """
    Flag as being public
    """
    def flag_public(self):
        self.public = True

    # -----------------------------------------------------------------------------
    # Support for GIR array/struct compat methods in 1.32

    """
    Emits the code to declare specific new types required by the variable
    """
    def emit_types_gir(self, hfile, cfile, since):
        pass

    """
    Builds the code to include the declaration of a variable of this kind,
    used when generating input/output bundles
    """
    def build_variable_declaration_gir(self, line_prefix, variable_name):
        # By default, no extra variable is required in the input/output bundle
        # for GIR compat methods
        return ''

    """
    Builds the code to include in the getter method declaration for this kind of variable
    """
    def build_getter_declaration_gir(self, line_prefix, variable_name):
        return self.build_getter_declaration(line_prefix, variable_name)

    """
    Builds the documentation of the getter code
    """
    def build_getter_documentation_gir(self, line_prefix, variable_name):
        return self.build_getter_documentation(line_prefix, variable_name)

    """
    Builds the code to implement getting this kind of variable
    """
    def build_getter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_getter_implementation(line_prefix, variable_name_from, variable_name_to)

    """
    Builds the code to include in the setter method declaration for this kind of variable
    """
    def build_setter_declaration_gir(self, line_prefix, variable_name):
        return self.build_setter_declaration(line_prefix, variable_name)

    """
    Builds the documentation of the setter code
    """
    def build_setter_documentation_gir(self, line_prefix, variable_name):
        return self.build_setter_documentation(line_prefix, variable_name)

    """
    Builds the code to implement setting this kind of variable
    """
    def build_setter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_setter_implementation(line_prefix, variable_name_from, variable_name_to)

    """
    Declaration of the variable as a struct field
    """
    def build_struct_field_declaration_gir(self, line_prefix, variable_name):
        return self.build_struct_field_declaration(line_prefix, variable_name)

    """
    Documentation for the struct field
    """
    def build_struct_field_documentation_gir(self, line_prefix, variable_name):
        return self.build_struct_field_documentation(line_prefix, variable_name)

    """
    Emits the code to dispose the variable
    """
    def build_dispose_gir(self, line_prefix, variable_name):
        return self.build_dispose(line_prefix, variable_name)

    """
    Emits the code to copy a variable into its GIR specific format
    """
    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        return ''

    """
    Emits the code to copy a variable from its GIR specific format
    """
    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        return ''

    """
    Emits the code to copy a variable from its GIR specific format to another
    variable in the same format
    """
    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        return ''



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils
from Variable import Variable
import VariableFactory

"""
Variable type for Structs ('struct' format)
These variables are exclusively used as array elements, they can not be used as
independent TLVs. Use Sequence instead for that.
"""
class VariableStruct(Variable):

    def __init__(self, service, dictionary, struct_type_name, container_type):

        # Call the parent constructor
        Variable.__init__(self, service, dictionary)

        self.container_type = container_type

        # The public format of the struct is built directly from the suggested
        # struct type name
        self.struct_type_name = struct_type_name
        self.element_type = utils.build_camelcase_name(self.struct_type_name)
        self.public_format = self.element_type
        self.private_format = self.public_format

        self.since = dictionary['since'] if 'since' in dictionary else ''

        # Load members of this struct
        self.members = []
        for member_dictionary in dictionary['contents']:
            member = {}
            member['name'] = utils.build_underscore_name(member_dictionary['name'])
            member['object'] = VariableFactory.create_variable(self.service, member_dictionary, struct_type_name + ' ' + member['name'], self.container_type)
            # Specify that the variable will be defined in the public header
            member['object'].flag_public()
            self.members.append(member)

        # We'll need to dispose if at least one of the members needs it
        for member in self.members:
            if member['object'].needs_dispose:
                self.needs_dispose = True
                break

        if self.needs_dispose:
            self.clear_method = '__' + utils.build_underscore_name(self.struct_type_name) + '_clear'


        # When using structs, we always require GIR compat methods because they're going
        # to be stored via reference pointers in the arrays
        self.needs_compat_gir = True

        # If any of the contents needs GIR compat methods, then we must also create a new
        # GIR compat struct type; otherwise just use the original one
        self.content_needs_compat_gir = False
        for member in self.members:
            if member['object'].needs_compat_gir:
                self.content_needs_compat_gir = True
                break

        if self.content_needs_compat_gir:
            self.struct_type_name_gir = self.struct_type_name + ' Gir'
        else:
            self.struct_type_name_gir = self.struct_type_name
        self.element_type_gir = utils.build_camelcase_name(self.struct_type_name_gir)
        self.public_format_gir = self.element_type_gir
        self.private_format_gir = self.public_format_gir
        self.new_method_gir = '__' + utils.build_underscore_name(self.struct_type_name_gir) + '_new'
        self.free_method_gir = '__' + utils.build_underscore_name(self.struct_type_name_gir) + '_free'

        # We'll contain personal info if at least one of the members contains personal info or we ourselves are personal info
        if not self.contains_personal_info:
            for member in self.members:
                if member['object'].contains_personal_info:
                    self.contains_personal_info = True
                    break


    def build_variable_declaration(self, line_prefix, variable_name):
        raise RuntimeError('Variable of type "struct" can only be defined as array members')


    def build_struct_field_declaration(self, line_prefix, variable_name):
        raise RuntimeError('Variable of type "struct" cannot be defined as struct fields')


    def emit_types(self, hfile, cfile, since, static):
        for member in self.members:
            member['object'].emit_types(hfile, cfile, since, static)

        translations = { 'format' : self.public_format,
                         'since'  : since if not self.since else self.since }
        template = '\n'
        hfile.write(string.Template(template).substitute(translations))

        if not static and self.service != 'CTL':
            template = (
                '\n'
                '/**\n')
            if self.content_needs_compat_gir:
                template += (
                    ' * ${format}: (skip)\n')
            else:
                template += (
                    ' * ${format}:\n')
            hfile.write(string.Template(template).substitute(translations))
            for member in self.members:
                hfile.write(member['object'].build_struct_field_documentation(' * ', member['name']))
            template = (
                ' *\n'
                ' * A ${format} struct.\n'
                ' *\n'
                ' * Since: ${since}\n'
                ' */\n')
            hfile.write(string.Template(template).substitute(translations))

        template = (
            'typedef struct _${format} {\n')
        hfile.write(string.Template(template).substitute(translations))

        for member in self.members:
            hfile.write(member['object'].build_struct_field_declaration('    ', member['name']))

        template = ('} ${format};\n')
        hfile.write(string.Template(template).substitute(translations))

        # No need for the clear func if no need to dispose the contents
        if self.needs_dispose:
            translations['clear_method'] = self.clear_method
            template = (
                '\n'
                'static void\n'
                '${clear_method} (${format} *value)\n'
                '{\n')
            for member in self.members:
                template += member['object'].build_dispose('    ', 'value->' + member['name'])
            template += (
                '}\n')
            cfile.write(string.Template(template).substitute(translations))


    def emit_types_gir(self, hfile, cfile, since):
        for member in self.members:
            member['object'].emit_types_gir(hfile, cfile, since)

        translations = { 'element_type_no_gir' : self.public_format,
                         'element_type'        : self.element_type_gir,
                         'underscore'          : utils.build_underscore_name(self.struct_type_name_gir),
                         'new_method'          : self.new_method_gir,
                         'free_method'         : self.free_method_gir,
                         'since'               : since if utils.version_compare('1.32',since) > 0 else '1.32' }

        # The type emission should happen ONLY if the GIR required type is
        # different to the original one
        if self.content_needs_compat_gir:
            template = (
                '\n'
                '/**\n'
                ' * ${element_type}: (rename-to ${element_type_no_gir})\n')
            hfile.write(string.Template(template).substitute(translations))
            for member in self.members:
                hfile.write(member['object'].build_struct_field_documentation_gir(' * ', member['name']))
            template = (
                ' *\n'
                ' * A ${element_type} struct.\n'
                ' *\n'
                ' * This type is a version of #${element_type_no_gir}, using arrays of pointers to\n'
                ' * structs instead of arrays of structs, for easier binding in other languages.\n'
                ' *\n'
                ' * Since: ${since}\n'
                ' */\n')
            hfile.write(string.Template(template).substitute(translations))

            template = (
                'typedef struct _${element_type} {\n')
            hfile.write(string.Template(template).substitute(translations))
            for member in self.members:
                hfile.write(member['object'].build_struct_field_declaration_gir('    ', member['name']))
                template = (
                    '} ${element_type};\n')
            hfile.write(string.Template(template).substitute(translations))

        # Free and New methods are required always for the GIR required type,
        # regardless of whether it's different to the original one or the same
        template = (
            '\n'
            'static void\n'
            '${free_method} (${element_type} *value)\n'
            '{\n')
        if not self.content_needs_compat_gir and self.needs_dispose:
            translations['clear_method'] = self.clear_method
            template += '    ${clear_method} (value);\n'
        else:
            for member in self.members:
                template += member['object'].build_dispose_gir('    ', 'value->' + member['name'])
        template += (
            '    g_slice_free (${element_type}, value);\n'
            '}\n'
            '\n'
            'static ${element_type} *\n'
            '${new_method} (void)\n'
            '{\n'
            '    return g_slice_new0 (${element_type});\n'
            '}\n')
        cfile.write(string.Template(template).substitute(translations))

        template = (
            '\n'
            'GType ${underscore}_get_type (void) G_GNUC_CONST;\n')
        hfile.write(string.Template(template).substitute(translations))

        template = (
            '\n'
            'static ${element_type} *\n'
            '__${underscore}_copy (const ${element_type} *value)\n'
            '{\n'
            '    ${element_type} *copy;\n'
            '\n'
            '    copy = ${new_method} ();\n')
        template += self.build_copy_gir('    ', 'value', 'copy')
        template += (
            '    return copy;\n'
            '}\n'
            '\n'
            'G_DEFINE_BOXED_TYPE (${element_type}, ${underscore}, (GBoxedCopyFunc)__${underscore}_copy, (GBoxedFreeFunc)__${underscore}_free)\n')
        cfile.write(string.Template(template).substitute(translations))


    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        for member in self.members:
            member['object'].emit_buffer_read(f, line_prefix, tlv_out, error, variable_name + '.' +  member['name'])


    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        for member in self.members:
            member['object'].emit_buffer_write(f, line_prefix, tlv_name, variable_name + '.' +  member['name'])


    def emit_get_printable(self, f, line_prefix, is_personal):
        translations = { 'lp' : line_prefix }

        template = (
            '${lp}g_string_append (printable, "[");\n')
        f.write(string.Template(template).substitute(translations))

        for member in self.members:
            translations['variable_name'] = member['name']
            template = (
                '${lp}g_string_append (printable, " ${variable_name} = \'");\n')
            f.write(string.Template(template).substitute(translations))

            member['object'].emit_get_printable(f, line_prefix, self.personal_info or is_personal)

            template = (
                '${lp}g_string_append (printable, "\'");\n')
            f.write(string.Template(template).substitute(translations))

        template = (
            '${lp}g_string_append (printable, " ]");\n')
        f.write(string.Template(template).substitute(translations))


    def build_dispose(self, line_prefix, variable_name):
        translations = { 'lp'            : line_prefix,
                         'underscore'    : utils.build_underscore_name(self.struct_type_name),
                         'variable_name' : variable_name }

        template = '${lp}${underscore}_clear (&${variable_name});\n'
        return string.Template(template).substitute(translations)


    def build_dispose_gir(self, line_prefix, variable_name):
        translations = { 'lp'            : line_prefix,
                         'underscore'    : utils.build_underscore_name(self.struct_type_name_gir),
                         'variable_name' : variable_name }

        template = '${lp}${underscore}_free (${variable_name});\n'
        return string.Template(template).substitute(translations)


    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'                 : line_prefix,
                         'element_type_gir'   : self.element_type_gir,
                         'variable_name_to'   : variable_name_to,
                         'variable_name_from' : variable_name_from }

        template = ''
        for member in self.members:
            template += member['object'].build_copy_to_gir(line_prefix,
                                                           '${variable_name_from}.' + member['name'],
                                                           '${variable_name_to}->' + member['name'])
        return string.Template(template).substitute(translations)


    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'                 : line_prefix,
                         'element_type_gir'   : self.element_type_gir,
                         'variable_name_to'   : variable_name_to,
                         'variable_name_from' : variable_name_from }

        template = ''
        for member in self.members:
            template += member['object'].build_copy_from_gir(line_prefix,
                                                             '${variable_name_from}->' + member['name'],
                                                             '${variable_name_to}.' + member['name'])
        return string.Template(template).substitute(translations)


    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'                 : line_prefix,
                         'element_type_gir'   : self.element_type_gir,
                         'variable_name_to'   : variable_name_to,
                         'variable_name_from' : variable_name_from }

        template = ''
        for member in self.members:
            template += member['object'].build_copy_gir(line_prefix,
                                                        '${variable_name_from}->' + member['name'],
                                                        '${variable_name_to}->' + member['name'])
        return string.Template(template).substitute(translations)


    def add_sections(self, sections):
        for member in self.members:
            member['object'].add_sections(sections)

        sections['public-types'] += self.element_type + '\n'
        sections['standard'] += utils.build_underscore_name(self.struct_type_name) + '_get_type\n'
        if self.content_needs_compat_gir:
            sections['public-types'] += self.element_type_gir + '\n'
            sections['standard'] += utils.build_underscore_name(self.struct_type_name_gir) + '_get_type\n'



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils
from Variable import Variable
import VariableFactory

"""
Variable type for Arrays ('array' format)
"""
class VariableArray(Variable):

    def __init__(self, service, dictionary, array_element_type, container_type):

        # Call the parent constructor
        Variable.__init__(self, service, dictionary)

        self.fixed_size = 0
        self.name = dictionary['name']

        # The array and its contents need to get disposed
        self.needs_dispose = True

        # We need to know whether the variable comes in an Input container or in
        # an Output container, as we should not dump the element clear() helper method
        # if the variable is from an Input container.
        self.container_type = container_type

        # Disallow arrays of arrays, always use an intermediate struct
        if dictionary['array-element']['format'] == 'array':
                raise ValueError('Arrays of arrays not allowed in %s array: use an intermediate struct instead' % self.name)

        # Load variable type of this array
        if 'name' in dictionary['array-element']:
            self.array_element = VariableFactory.create_variable(self.service,
                                                                 dictionary['array-element'],
                                                                 array_element_type + ' ' + dictionary['array-element']['name'],
                                                                 self.container_type)
        else:
            self.array_element = VariableFactory.create_variable(self.service,
                                                                 dictionary['array-element'],
                                                                 '',
                                                                 self.container_type)

        if 'size-prefix-format' in dictionary and 'fixed-size' in dictionary:
            raise ValueError('Cannot give \'size-prefix-format\' and \'fixed-size\' in %s array at the same time' % self.name)
        elif 'size-prefix-format' in dictionary:
            # Load variable type for the array size prefix
            # We do NOT allow 64-bit types as array sizes (GArray won't support them)
            if dictionary['size-prefix-format'] not in [ 'guint8', 'guint16', 'guint32' ]:
                raise ValueError('Invalid size prefix format (%s) in %s array: not guint8 or guint16 or guint32' %
                                 (dictionary['size-prefix-format'], self.name))
            default_array_size = { 'format' : dictionary['size-prefix-format'] }
            self.array_size_element = VariableFactory.create_variable(self.service, default_array_size, '', self.container_type)
            # Load variable type for the sequence prefix, if any
            if 'sequence-prefix-format' in dictionary:
                sequence = { 'format' : dictionary['sequence-prefix-format'] }
                self.array_sequence_element = VariableFactory.create_variable(self.service, sequence, '', self.container_type)
            else:
                self.array_sequence_element = ''

        elif 'fixed-size' in dictionary:
            # fixed-size arrays have no size element, obviously
            self.fixed_size = dictionary['fixed-size']
            if int(self.fixed_size) == 0 or int(self.fixed_size) > 512:
                raise ValueError('Fixed array size %s out of bounds (not between 0 and 512)' % self.fixed_size)
            if 'sequence-prefix-format' in dictionary:
                raise ValueError('\'sequence-prefix-format\' is not supported along with \'fixed-size\' in %s array' % self.name)
            else:
                self.array_sequence_element = ''
        else:
            raise ValueError('Missing \'size-prefix-format\' or \'fixed-size\' in %s array' % self.name)

        # Arrays need compat GIR support if the array element needs compat GIR support
        self.needs_compat_gir = self.array_element.needs_compat_gir

        # Arrays contain personal info if the array element contains personal info or they themselves are personal info
        self.contains_personal_info = self.contains_personal_info or self.array_element.contains_personal_info


    def emit_types(self, hfile, cfile, since, static):
        self.array_element.emit_types(hfile, cfile, since, static)


    def emit_types_gir(self, hfile, cfile, since):
        if self.array_element.needs_compat_gir:
            self.array_element.emit_types_gir(hfile, cfile, since)


    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'                          : line_prefix,
                         'variable_name'               : variable_name,
                         'private_format'              : self.private_format,
                         'array_element_public_format' : self.array_element.public_format,
                         'array_element_clear_method'  : self.array_element.clear_method,
                         'common_var_prefix'           : common_var_prefix }

        template = (
            '${lp}{\n'
            '${lp}    guint ${common_var_prefix}_i;\n')
        f.write(string.Template(template).substitute(translations))

        if self.fixed_size:
            translations['fixed_size'] = self.fixed_size

            template = (
                '${lp}    guint16 ${common_var_prefix}_n_items = ${fixed_size};\n'
                '\n')
            f.write(string.Template(template).substitute(translations))
        else:
            translations['array_size_element_format'] = self.array_size_element.public_format
            template = (
                '${lp}    ${array_size_element_format} ${common_var_prefix}_n_items;\n')

            if self.array_sequence_element != '':
                translations['array_sequence_element_format'] = self.array_sequence_element.public_format
                template += (
                    '${lp}    ${array_sequence_element_format} ${common_var_prefix}_sequence;\n')

            template += (
                '\n'
                '${lp}    /* Read number of items in the array */\n')
            f.write(string.Template(template).substitute(translations))
            self.array_size_element.emit_buffer_read(f, line_prefix + '    ', tlv_out, error, common_var_prefix + '_n_items')

            if self.array_sequence_element != '':
                template = (
                    '\n'
                    '${lp}    /* Read sequence in the array */\n')
                f.write(string.Template(template).substitute(translations))
                self.array_size_element.emit_buffer_read(f, line_prefix + '    ', tlv_out, error, common_var_prefix + '_sequence')

                template = (
                    '\n'
                    '${lp}    ${variable_name}_sequence = ${common_var_prefix}_sequence;\n')
                f.write(string.Template(template).substitute(translations))

        template = (
            '\n'
            '${lp}    ${variable_name} = g_array_sized_new (\n'
            '${lp}        FALSE,\n'
            '${lp}        FALSE,\n'
            '${lp}        sizeof (${array_element_public_format}),\n'
            '${lp}        (guint)${common_var_prefix}_n_items);\n'
            '\n')

        if self.array_element.needs_dispose:
            template += (
                '${lp}    g_array_set_clear_func (${variable_name}, (GDestroyNotify)${array_element_clear_method});\n'
                '\n')

        template += (
            '${lp}    for (${common_var_prefix}_i = 0; ${common_var_prefix}_i < ${common_var_prefix}_n_items; ${common_var_prefix}_i++) {\n'
            '${lp}        ${array_element_public_format} ${common_var_prefix}_aux;\n'
            '\n')
        f.write(string.Template(template).substitute(translations))

        self.array_element.emit_buffer_read(f, line_prefix + '        ', tlv_out, error, common_var_prefix + '_aux')

        template = (
            '${lp}        g_array_insert_val (${variable_name}, ${common_var_prefix}_i, ${common_var_prefix}_aux);\n'
            '${lp}    }\n'
            '${lp}}\n')
        f.write(string.Template(template).substitute(translations))


    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'                : line_prefix,
                         'variable_name'     : variable_name,
                         'common_var_prefix' : common_var_prefix }

        template = (
            '${lp}{\n'
            '${lp}    guint ${common_var_prefix}_i;\n')
        f.write(string.Template(template).substitute(translations))

        if self.fixed_size == 0:
            translations['array_size_element_format'] = self.array_size_element.private_format

            template = (
                '${lp}    ${array_size_element_format} ${common_var_prefix}_n_items;\n'
                '\n'
                '${lp}    /* Write the number of items in the array first */\n'
                '${lp}    ${common_var_prefix}_n_items = (${array_size_element_format}) ${variable_name}->len;\n')
            f.write(string.Template(template).substitute(translations))

            self.array_size_element.emit_buffer_write(f, line_prefix + '    ', tlv_name, common_var_prefix + '_n_items')

        if self.array_sequence_element != '':
            self.array_sequence_element.emit_buffer_write(f, line_prefix + '    ', tlv_name, variable_name + '_sequence')


        template = (
            '\n'
            '${lp}    for (${common_var_prefix}_i = 0; ${common_var_prefix}_i < ${variable_name}->len; ${common_var_prefix}_i++) {\n')
        f.write(string.Template(template).substitute(translations))

        self.array_element.emit_buffer_write(f, line_prefix + '        ', tlv_name, 'g_array_index (' + variable_name + ', ' + self.array_element.public_format + ',' + common_var_prefix + '_i)')

        template = (
            '${lp}    }\n'
            '${lp}}\n')
        f.write(string.Template(template).substitute(translations))


    def emit_get_printable(self, f, line_prefix, is_personal):
        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'                : line_prefix,
                         'common_var_prefix' : common_var_prefix }

        template = (
            '${lp}{\n'
            '${lp}    guint ${common_var_prefix}_i;\n')
        f.write(string.Template(template).substitute(translations))

        if self.fixed_size:
            translations['fixed_size'] = self.fixed_size

            template = (
                '${lp}    guint16 ${common_var_prefix}_n_items = ${fixed_size};\n'
                '\n')
            f.write(string.Template(template).substitute(translations))
        else:
            translations['array_size_element_format'] = self.array_size_element.public_format
            template = (
                '${lp}    ${array_size_element_format} ${common_var_prefix}_n_items;\n')

            if self.array_sequence_element != '':
                translations['array_sequence_element_format'] = self.array_sequence_element.public_format
                template += (
                    '${lp}    ${array_sequence_element_format} ${common_var_prefix}_sequence;\n')

            template += (
                '\n'
                '${lp}    /* Read number of items in the array */\n')
            f.write(string.Template(template).substitute(translations))
            self.array_size_element.emit_buffer_read(f, line_prefix + '    ', 'out', '&error', common_var_prefix + '_n_items')

            if self.array_sequence_element != '':
                template = (
                    '\n'
                    '${lp}    /* Read sequence */\n')
                f.write(string.Template(template).substitute(translations))
                self.array_sequence_element.emit_buffer_read(f, line_prefix + '    ', 'out', '&error', common_var_prefix + '_sequence')
                template = (
                    '\n'
                    '${lp}    g_string_append_printf (printable, "[[Seq:%u]] ", ${common_var_prefix}_sequence);\n')
                f.write(string.Template(template).substitute(translations))

        template = (
            '\n'
            '${lp}    g_string_append (printable, "{");\n'
            '\n'
            '${lp}    for (${common_var_prefix}_i = 0; ${common_var_prefix}_i < ${common_var_prefix}_n_items; ${common_var_prefix}_i++) {\n'
            '${lp}        g_string_append_printf (printable, " [%u] = \'", ${common_var_prefix}_i);\n')
        f.write(string.Template(template).substitute(translations))

        self.array_element.emit_get_printable(f, line_prefix + '        ', self.personal_info or is_personal);

        template = (
            '${lp}        g_string_append (printable, " \'");\n'
            '${lp}    }\n'
            '\n'
            '${lp}    g_string_append (printable, "}");\n'
            '${lp}}')
        f.write(string.Template(template).substitute(translations))


    """
    We need to include SEQUENCE + GARRAY
    """
    def build_variable_declaration(self, line_prefix, variable_name):
        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''
        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} ${name}_sequence;\n')

        template += (
            '${lp}GArray *${name};\n')
        return string.Template(template).substitute(translations)


    """
    We need to include GPTRARRAY
    """
    def build_variable_declaration_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return ''

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = (
            '${lp}GPtrArray *${name}_ptr;\n')
        return string.Template(template).substitute(translations)


    """
    We need to include SEQUENCE + GARRAY
    """
    def build_struct_field_declaration(self, line_prefix, variable_name):
        return self.build_variable_declaration(line_prefix, variable_name)


    """
    We need to include SEQUENCE + GPTRARRAY
    """
    def build_struct_field_declaration_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_struct_field_declaration(line_prefix, variable_name)

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''

        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} ${name}_sequence;\n')

        template += (
            '${lp}GPtrArray *${name};\n')
        return string.Template(template).substitute(translations)


    def build_getter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ''

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''
        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} *${name}_sequence,\n')

        template += (
            '${lp}GArray **${name},\n')
        return string.Template(template).substitute(translations)


    def build_getter_declaration_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_getter_declaration(line_prefix, variable_name)

        if not self.visible:
            return ''

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''
        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} *${name}_sequence,\n')

        template += (
            '${lp}GPtrArray **${name}_ptr,\n')
        return string.Template(template).substitute(translations)


    def build_getter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ''

        translations = { 'lp'                          : line_prefix,
                         'array_element_public_format' : self.array_element.public_format,
                         'array_element_element_type'  : self.array_element.element_type,
                         'name'                        : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: (out)(optional): a placeholder for the output sequence number, or %NULL if not required.\n')

        template += (
            '${lp}@${name}: (out)(optional)(element-type ${array_element_element_type})(transfer none): a placeholder for the output #GArray of #${array_element_public_format} elements, or %NULL if not required. Do not free it, it is owned by @self.\n')
        return string.Template(template).substitute(translations)


    def build_getter_documentation_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_getter_documentation(line_prefix, variable_name)
        if not self.visible:
            return ''

        translations = { 'lp'                             : line_prefix,
                         'array_element_public_format'    : self.array_element.public_format,
                         'array_element_element_type_gir' : self.array_element.element_type_gir,
                         'name'                           : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: (out)(optional): a placeholder for the output sequence number, or %NULL if not required.\n')

        template += (
            '${lp}@${name}_ptr: (out)(optional)(element-type ${array_element_element_type_gir})(transfer none): a placeholder for the output array of #${array_element_public_format} elements, or %NULL if not required. Do not free or modify it, it is owned by @self.\n')
        return string.Template(template).substitute(translations)


    def build_getter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ''

        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}if (${to}_sequence)\n'
                '${lp}    *${to}_sequence = ${from}_sequence;\n')

        template += (
            '${lp}if (${to})\n'
            '${lp}    *${to} = ${from};\n')

        return string.Template(template).substitute(translations)


    def build_getter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        if not self.array_element.needs_compat_gir:
            return self.build_getter_implementation(line_prefix, variable_name_from, variable_name_to)
        if not self.visible:
            return ''

        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}if (${to}_sequence)\n'
                '${lp}    *${to}_sequence = ${from}_sequence;\n')

        template += (
            '${lp}if (${to}_ptr) {\n'
            '${lp}    if (!${from}_ptr) {\n')

        template += self.build_copy_to_gir(line_prefix + '        ',
                                           variable_name_from,
                                           variable_name_from + '_ptr')

        template += (
            '${lp}    }\n'
            '${lp}    *${to}_ptr = ${from}_ptr;\n'
            '${lp}}')

        return string.Template(template).substitute(translations)


    def build_setter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ''

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''
        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} ${name}_sequence,\n')

        template += (
            '${lp}GArray *${name},\n')
        return string.Template(template).substitute(translations)


    def build_setter_declaration_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_setter_declaration(line_prefix, variable_name)
        if not self.visible:
            return ''

        translations = { 'lp'   : line_prefix,
                         'name' : variable_name }

        template = ''
        if self.array_sequence_element != '':
            translations['array_sequence_element_format'] = self.array_sequence_element.public_format
            template += (
                '${lp}${array_sequence_element_format} ${name}_sequence,\n')

        template += (
            '${lp}GPtrArray *${name}_ptr,\n')
        return string.Template(template).substitute(translations)


    def build_setter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'                          : line_prefix,
                         'array_element_public_format' : self.array_element.public_format,
                         'array_element_element_type'  : self.array_element.element_type,
                         'name'                        : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: the sequence number.\n')

        template += (
            '${lp}@${name}: (in)(element-type ${array_element_element_type})(transfer none): a #GArray of #${array_element_public_format} elements. A new reference to @${name} will be taken, so the caller must make sure the array was created with the correct #GDestroyNotify as clear function for each element in the array.\n')
        return string.Template(template).substitute(translations)


    def build_setter_documentation_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_setter_documentation(line_prefix, variable_name)
        if not self.visible:
            return ""

        translations = { 'lp'                             : line_prefix,
                         'array_element_public_format'    : self.array_element.public_format,
                         'array_element_element_type_gir' : self.array_element.element_type_gir,
                         'name'                           : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: the sequence number.\n')

        template += (
            '${lp}@${name}_ptr: (in)(element-type ${array_element_element_type_gir})(transfer none): array of #${array_element_public_format} elements. The contents of the given array will be copied, the #GPtrArray will not increase its reference count.\n')
        return string.Template(template).substitute(translations)


    def build_setter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}${to}_sequence = ${from}_sequence;\n')

        template += self.build_dispose(line_prefix, variable_name_to)
        template += self.build_dispose_gir(line_prefix, variable_name_to)

        template +=(
            '${lp}${to} = g_array_ref (${from});\n')
        return string.Template(template).substitute(translations)


    def build_setter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        if not self.array_element.needs_compat_gir:
            return self.build_setter_implementation(line_prefix, variable_name_from, variable_name_to)
        if not self.visible:
            return ""

        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'   : line_prefix,
                         'from' : variable_name_from,
                         'to'   : variable_name_to }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}${to}_sequence = ${from}_sequence;\n')

        template += self.build_dispose(line_prefix, variable_name_to)
        template += self.build_dispose_gir(line_prefix, variable_name_to + '_ptr')
        template += self.build_copy_from_gir(line_prefix, variable_name_from + '_ptr', variable_name_to)

        return string.Template(template).substitute(translations)


    def build_struct_field_documentation(self, line_prefix, variable_name):
        translations = { 'lp'                          : line_prefix,
                         'array_element_public_format' : self.array_element.public_format,
                         'name'                        : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: the sequence number.\n')

        template += (
            '${lp}@${name}: a #GArray of #${array_element_public_format} elements.\n')
        return string.Template(template).substitute(translations)


    def build_struct_field_documentation_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            return self.build_struct_field_documentation(line_prefix, variable_name)

        translations = { 'lp'                          : line_prefix,
                         'array_element_public_format' : self.array_element.public_format,
                         'array_element_element_type'  : self.array_element.element_type,
                         'name'                        : variable_name }

        template = ''
        if self.array_sequence_element != '':
            template += (
                '${lp}@${name}_sequence: the sequence number.\n')

        template += (
            '${lp}@${name}: (element-type ${array_element_element_type}): an array of #${array_element_public_format} elements.\n')
        return string.Template(template).substitute(translations)


    def build_dispose(self, line_prefix, variable_name):
        translations = { 'lp'            : line_prefix,
                         'variable_name' : variable_name }

        template = (
            '${lp}g_clear_pointer (&${variable_name}, (GDestroyNotify)g_array_unref);\n')
        return string.Template(template).substitute(translations)


    def build_dispose_gir(self, line_prefix, variable_name):
        if not self.array_element.needs_compat_gir:
            self.build_dispose(line_prefix, variable_name)

        translations = { 'lp'            : line_prefix,
                         'variable_name' : variable_name }

        template = (
            '${lp}g_clear_pointer (&${variable_name}, (GDestroyNotify)g_ptr_array_unref);\n')
        return string.Template(template).substitute(translations)


    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'   : line_prefix,
                         'to'   : variable_name_to,
                         'from' : variable_name_from }

        if self.array_element.needs_compat_gir:

            translations['common_var_prefix']               = common_var_prefix
            translations['array_element_public_format']     = self.array_element.public_format
            translations['array_element_public_format_gir'] = self.array_element.public_format_gir
            translations['array_element_element_type_gir']  = self.array_element.element_type_gir
            translations['array_element_free_method_gir']   = self.array_element.free_method_gir
            translations['array_element_new_method_gir']    = self.array_element.new_method_gir

            template = (
                '${lp}{\n'
                '${lp}    guint ${common_var_prefix}_i;\n'
                '\n'
                '${lp}    ${to} = g_ptr_array_new_full (${from}->len, (GDestroyNotify)${array_element_free_method_gir});\n'
                '${lp}    for (${common_var_prefix}_i = 0; ${common_var_prefix}_i < ${from}->len; ${common_var_prefix}_i++) {\n'
                '${lp}        ${array_element_public_format} *${common_var_prefix}_aux_from;\n'
                '${lp}        ${array_element_public_format_gir} *${common_var_prefix}_aux_to;\n'
                '\n'
                '${lp}        ${common_var_prefix}_aux_from = &g_array_index (${from}, ${array_element_public_format}, ${common_var_prefix}_i);\n'
                '\n'
                '${lp}        ${common_var_prefix}_aux_to = ${array_element_new_method_gir} ();\n')

            template += self.array_element.build_copy_to_gir(line_prefix + '        ',
                                                             '(*${common_var_prefix}_aux_from)',
                                                             '${common_var_prefix}_aux_to')

            template += (
                '\n'
                '${lp}        g_ptr_array_add (${to}, ${common_var_prefix}_aux_to);\n'
                '${lp}    }\n'
                '${lp}}\n')
        else:
            template = (
                '${lp}${to} = g_array_ref (${from});\n')
        return string.Template(template).substitute(translations)


    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'   : line_prefix,
                         'to'   : variable_name_to,
                         'from' : variable_name_from }

        if self.array_element.needs_compat_gir:
            common_var_prefix = utils.build_underscore_name(self.name)
            translations['common_var_prefix']               = common_var_prefix
            translations['array_element_public_format']     = self.array_element.public_format
            translations['array_element_public_format_gir'] = self.array_element.public_format_gir

            template = (
                '${lp}{\n'
                '${lp}    guint ${common_var_prefix}_i;\n'
                '\n'
                '${lp}    ${to} = g_array_sized_new (FALSE, FALSE, sizeof (${array_element_public_format}), ${from}->len);\n'
                '${lp}    for (${common_var_prefix}_i = 0; ${common_var_prefix}_i < ${from}->len; ${common_var_prefix}_i++) {\n'
                '${lp}        ${array_element_public_format} ${common_var_prefix}_aux_to;\n'
                '${lp}        ${array_element_public_format_gir} *${common_var_prefix}_aux_from;\n'
                '\n'
                '${lp}        ${common_var_prefix}_aux_from = g_ptr_array_index (${from}, ${common_var_prefix}_i);\n'
                '\n')

            template += self.array_element.build_copy_from_gir(line_prefix + '        ',
                                                               '${common_var_prefix}_aux_from',
                                                               '${common_var_prefix}_aux_to')

            template += (
                '\n'
                '${lp}        g_array_append_val (${to}, ${common_var_prefix}_aux_to);\n'
                '${lp}    }\n'
                '${lp}}\n')
        else:
            template = (
                '${lp}${to} = g_array_ref (${from});\n')
        return string.Template(template).substitute(translations)


    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        common_var_prefix = utils.build_underscore_name(self.name)
        translations = { 'lp'   : line_prefix,
                         'to'   : variable_name_to,
                         'from' : variable_name_from }

        # our GPtrArrays and GArrays are immutable, we can just ref the array as a copy
        if self.array_element.needs_compat_gir:
            template = (
                '${lp}${to} = g_ptr_array_ref (${from});\n')
        else:
            template = (
                '${lp}${to} = g_array_ref (${from});\n')
        return string.Template(template).substitute(translations)


    def add_sections(self, sections):
        self.array_element.add_sections(sections)



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils
from Variable import Variable
import VariableFactory

"""
Variable type for Sequences ('sequence' format)
"""
class VariableSequence(Variable):

    def __init__(self, service, dictionary, sequence_type_name, container_type):

        # Call the parent constructor
        Variable.__init__(self, service, dictionary)

        self.container_type = container_type

        # Load members of this sequence
        self.members = []
        for member_dictionary in dictionary['contents']:
            member = {}
            member['name'] = utils.build_underscore_name(member_dictionary['name'])
            member['object'] = VariableFactory.create_variable(self.service, member_dictionary, sequence_type_name + ' ' + member_dictionary['name'], self.container_type)
            self.members.append(member)

        # We'll need to dispose if at least one of the members needs it
        for member in self.members:
            if member['object'].needs_dispose:
                self.needs_dispose = True
                break

        # We'll flag the variable as needing compat GIR methods if we find any
        # sequence member needing it
        for member in self.members:
            if member['object'].needs_compat_gir:
                self.needs_compat_gir = True
                break

        # We'll contain personal info if at least one of the members contains personal info or we ourselves are personal info
        if not self.contains_personal_info:
            for member in self.members:
                if member['object'].contains_personal_info:
                    self.contains_personal_info = True
                    break


    def emit_types(self, hfile, cfile, since, static):
        for member in self.members:
            member['object'].emit_types(hfile, cfile, since, static)


    def emit_types_gir(self, hfile, cfile, since):
        for member in self.members:
            member['object'].emit_types_gir(hfile, cfile, since)


    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        for member in self.members:
            member['object'].emit_buffer_read(f, line_prefix, tlv_out, error, variable_name + '_' +  member['name'])


    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        for member in self.members:
            member['object'].emit_buffer_write(f, line_prefix, tlv_name, variable_name + '_' +  member['name'])


    def emit_get_printable(self, f, line_prefix, is_personal):
        translations = { 'lp' : line_prefix }

        template = (
            '${lp}g_string_append (printable, "[");\n')
        f.write(string.Template(template).substitute(translations))

        for member in self.members:
            translations['variable_name'] = member['name']
            template = (
                '${lp}g_string_append (printable, " ${variable_name} = \'");\n')
            f.write(string.Template(template).substitute(translations))

            member['object'].emit_get_printable(f, line_prefix, self.personal_info or is_personal)

            template = (
                '${lp}g_string_append (printable, "\'");\n')
            f.write(string.Template(template).substitute(translations))

        template = (
            '${lp}g_string_append (printable, " ]");\n')
        f.write(string.Template(template).substitute(translations))


    def build_variable_declaration(self, line_prefix, variable_name):
        built = ''
        for member in self.members:
            built += member['object'].build_variable_declaration(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_variable_declaration_gir(self, line_prefix, variable_name):
        built = ''
        for member in self.members:
            built += member['object'].build_variable_declaration_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_struct_field_declaration(self, line_prefix, variable_name):
        raise RuntimeError('Variable of type "sequence" is never expected as a struct field')


    def build_getter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_declaration(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_getter_declaration_gir(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_declaration_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_getter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_documentation(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_getter_documentation_gir(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_documentation_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_getter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_implementation(line_prefix,
                                                                  variable_name_from + '_' + member['name'],
                                                                  variable_name_to + '_' + member['name'])
        return built


    def build_getter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_getter_implementation_gir(line_prefix,
                                                                      variable_name_from + '_' + member['name'],
                                                                      variable_name_to + '_' + member['name'])
        return built


    def build_setter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_declaration(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_setter_declaration_gir(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_declaration_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_setter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_documentation(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_setter_documentation_gir(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_documentation_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_setter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_implementation(line_prefix,
                                                                  variable_name_from + '_' + member['name'],
                                                                  variable_name_to + '_' + member['name'])
        return built


    def build_setter_implementation_gir(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        built = ''
        for member in self.members:
            built += member['object'].build_setter_implementation_gir(line_prefix,
                                                                      variable_name_from + '_' + member['name'],
                                                                      variable_name_to + '_' + member['name'])
        return built


    def build_dispose(self, line_prefix, variable_name):
        built = ''
        for member in self.members:
            built += member['object'].build_dispose(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_dispose_gir(self, line_prefix, variable_name):
        built = ''
        for member in self.members:
            built += member['object'].build_dispose_gir(line_prefix, variable_name + '_' + member['name'])
        return built


    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        built = ''
        for member in self.members:
            built += member['object'].build_copy_to_gir(line_prefix, variable_name_from, variable_name_to)
        return built


    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        built = ''
        for member in self.members:
            built += member['object'].build_copy_from_gir(line_prefix, variable_name_from, variable_name_to)
        return built


    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        built = ''
        for member in self.members:
            built += member['object'].build_copy_gir(line_prefix, variable_name_from, variable_name_to)
        return built


    def add_sections(self, sections):
        # Add sections for each member
        for member in self.members:
            member['object'].add_sections(sections)



# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import utils


# eDB2FieldType
FIELD_TYPE_STD           = 0 # Field is a standard type (see below)
FIELD_TYPE_ENUM_UNSIGNED = 1 # Field is an unsigned enumerated type
FIELD_TYPE_ENUM_SIGNED   = 2 # Field is a signed enumerated type

# eDB2StdFieldType
FIELD_STD_BOOL = 0         # boolean (0/1, false/true)
FIELD_STD_INT8 = 1         # 8-bit signed integer
FIELD_STD_UINT8 = 2        # 8-bit unsigned integer
FIELD_STD_INT16 = 3        # 16-bit signed integer
FIELD_STD_UINT16 = 4       # 16-bit unsigned integer
FIELD_STD_INT32 = 5        # 32-bit signed integer
FIELD_STD_UINT32 = 6       # 32-bit unsigned integer
FIELD_STD_INT64 = 7        # 64-bit signed integer
FIELD_STD_UINT64 = 8       # 64-bit unsigned integer
FIELD_STD_STRING_A = 9     # ANSI (ASCII?) fixed length string; size in bits
FIELD_STD_STRING_U = 10    # UCS-2 fixed length string
FIELD_STD_STRING_ANT = 11  # ANSI (ASCII?) NULL terminated string
FIELD_STD_STRING_UNT = 12  # UCS-2 NULL terminated string
FIELD_STD_FLOAT32 = 13     # 32-bit floating point value
FIELD_STD_FLOAT64 = 14     # 64-bit floating point value
FIELD_STD_STRING_U8 = 15   # UTF-8 encoded fixed length string
FIELD_STD_STRING_U8NT = 16 # UTF-8 encoded NULL terminated string

stdtypes = {
    # Maps field type to [ <C type>, <is array>, <element size in bits> ]
    FIELD_STD_BOOL: [ 'bool', False, 8 ],
    FIELD_STD_INT8: [ 'int8', False, 8 ],
    FIELD_STD_UINT8: [ 'uint8', False, 8 ],
    FIELD_STD_INT16: [ 'int16', False, 16 ],
    FIELD_STD_UINT16: [ 'uint16', False, 16 ],
    FIELD_STD_INT32: [ 'int32', False, 32 ],
    FIELD_STD_UINT32: [ 'uint32', False, 32 ],
    FIELD_STD_INT64: [ 'int64', False, 64 ],
    FIELD_STD_UINT64: [ 'uint64', False, 64 ],
    FIELD_STD_STRING_A: [ 'char', True, 8 ],
    FIELD_STD_STRING_U: [ 'uint16', True, 16 ],
    FIELD_STD_STRING_ANT: [ 'char *', False, 8 ],
    FIELD_STD_STRING_UNT: [ 'char *', False, 8 ],
    FIELD_STD_FLOAT32: [ 'float32', False, 32 ],
    FIELD_STD_FLOAT64: [ 'float64', False, 64 ],
}

def is_array_type(t):
    return stdtypes[t][1]


class Field:
    def __init__(self, line):
        parts = line.split('^')
        if len(parts) < 6:
            raise Exception("Invalid field line '%s'" % line)
        self.id = int(parts[0])
        self.name = parts[1].replace('"', '')
        self.size = int(parts[2]) # in *bits*
        self.type = int(parts[3]) # eDB2FieldType
        self.typeval = int(parts[4])  # eDB2StdFieldType if 'type' == 0
        self.hex = int(parts[5]) != 0
        self.descid = None
        self.internal = False
        if len(parts) > 6:
            self.descid = int(parts[6])
        if len(parts) > 7:
            self.internal = int(parts[7])

        # Field.txt:50118^"IP V4 Address"^8^0^2^0

    #Field.txt:50118^"IP V4 Address"^8^0^2^0

    def get_charsize(self):
        if self.type == FIELD_TYPE_STD:
            if self.typeval == FIELD_STD_STRING_A or self.typeval == FIELD_STD_STRING_ANT or \
                self.typeval == FIELD_STD_STRING_U or self.typeval == FIELD_STD_STRING_UNT:
                return stdtypes[self.typeval][2]
        raise Exception("Called for non-string type")

    def emit(self, do_print, indent, enums, num_elements, comment, isarray):
        ctype = ''
        arraypart = ''

        sizebits = 0
        if self.type == FIELD_TYPE_STD:  # eDB2_FIELD_STD
            tinfo = stdtypes[self.typeval]
            ctype = tinfo[0]
            
            if is_array_type(self.typeval) or num_elements > 0:
                if num_elements > 0:
                    arraypart = "[%d]" % num_elements
                    sizebits = num_elements * tinfo[2]
                elif isarray:
                    # array with size given by previous fragment
                    arraypart = "[0]"
                    sizebits = 0
                else:
                    arraypart = "[%d]" % (self.size / tinfo[2])
                    sizebits = self.size * tinfo[2]
        elif self.type == FIELD_TYPE_ENUM_UNSIGNED or self.type == FIELD_TYPE_ENUM_SIGNED:
            # It's a enum; find the enum
            e = enums.get_child(self.typeval)
            ctype = "gobi_%s" % utils.nicename(e.name)
            if isarray:
                if num_elements != 0:
                    raise Exception("Unhandled ENUM field type with size %d" % num_elements)
                arraypart = "[0]";
                sizebits = 0
            else:
                if self.size > 0:
                    # enum size is # of bits
                    arraypart = ":%d" % self.size
                    sizebits = self.size
                else:
                    sizebits = 32
        else:
            raise ValueError("Unknown Field type")

        if comment:
            comment = " (%s)" % comment
        if do_print:
            print "%s%s %s%s; /* %s%s */" % ("\t" * indent, ctype, utils.nicename(self.name), arraypart, self.name, comment)
        return sizebits

class Fields:
    def __init__(self, path):
        self.byid = {}

        f = file(path + "Field.txt")
        for line in f:
            field = Field(line.strip())
            self.byid[field.id] = field

    def has_child(self, fid):
        return fid in self.byid

    def get_child(self, fid):
        return self.byid[fid]




# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

def constname(name):
    rlist = { '(': '',
              ')': '',
              ' ': '_',
              '/': '_',
              '.': '',
              ',': '' }
    # Capture stuff like "-9 dB" correctly
    if name.find("dB") == -1:
        rlist['-'] = '_'
    else:
        rlist['-'] = 'NEG_'

    sanitized = ""
    for c in name:
        try:
            sanitized += rlist[c]
        except KeyError:
            sanitized += c

    # Handle E_UTRA -> EUTRA and E_UTRAN -> EUTRAN
    foo = sanitized.upper().strip('_')
    foo.replace('E_UTRA', 'EUTRA')

    # And HSDPA+ -> HSDPA_PLUS then strip all plusses
    foo = foo.replace('HSDPA+', 'HSDPA_PLUS')
    foo = foo.replace('+', '')

    # 1xEV-DO -> 1x EVDO
    foo = foo.replace("1XEV_DO", "1X_EVDO")
    foo = foo.replace("EV_DO", "EVDO")

    # Wi-Fi -> WiFi
    foo = foo.replace("WI_FI", "WIFI")

    # modify certain words
    words = foo.split('_')
    klist = [ 'UNSUPPORTED' ]
    slist = [ 'STATES', 'REASONS', 'FORMATS', 'SELECTIONS', 'TYPES', 'TAGS',
              'PROTOCOLS', 'CLASSES', 'ACTIONS', 'VALUES', 'OPTIONS',
              'DOMAINS', 'DEVICES', 'MODES', 'MONTHS', 'PREFERENCES',
              'PREFS', 'DURATIONS', 'CAPABILITIES', 'INTERFACES',
              'TECHNOLOGIES', 'NETWORKS', 'DAYS', 'SYSTEMS', 'SCHEMES',
              'INDICATORS', 'ENCODINGS', 'INITIALS', 'BITS', 'MEDIUMS',
              'BASES', 'ERRORS', 'RESULTS', 'RATIOS', 'DELIVERIES',
              'FAMILIES', 'SETTINGS', 'SOURCES', 'ORDERS' ]
    blist = { 'CLASSES': 'CLASS' }
    final = ''
    for word in words:
        if word in klist or len(word) == 0:
            continue
        if len(final):
            final += '_'
        if word in blist:
            final += blist[word]
        elif word in slist:
            if word.endswith("IES"):
                final += word[:len(word) - 3] + "Y"
            elif word.endswith("S"):
                final += word[:len(word) - 1]
        else:
            final += word
    return final

def nicename(name):
    name = name.lower()
    name = name.replace("1xev-do", "1x evdo")
    name = name.replace("ev-do", "evdo")
    name = name.replace("wi-fi", "wifi")
    name = name.replace("%", "pct")
    name = name.replace(' ', '_').replace('/', '_').replace('-', '_').replace('.','').replace('(', '').replace(')', '')
    name = name.replace("___", "_").replace("__", "_")
    return name.strip('_')


class DbFile:
    # Base class for objects that handle reading a database file like
    # Enum.txt or Struct.txt

    def __init__(self, path):
        raise Exception("init() method must be implemented")

    def validate(self):
        pass

    def has_child(self, cid):
        raise Exception("has_child() method must be implemented")

    def get_child(self, cid):
        raise Exception("get_child() method must be implemented")

    def emit(self):
        pass

    def emit_unused(self, used, fields, enums):
        pass




#!/usr/bin/env python
# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import sys
import Entities
import Enums
import Fields
import Structs

if len(sys.argv) > 2:
    print "Usage: qmidb.py <path to Entity.txt>"
    sys.exit(1)
path = ""
if len(sys.argv) == 2:
    path = sys.argv[1] + "/"

enums = Enums.Enums(path)
entities = Entities.Entities(path)
fields = Fields.Fields(path)
structs = Structs.Structs(path)

structs.validate(fields)
entities.validate(structs)

print '/* GENERATED CODE. DO NOT EDIT. */'
print '\ntypedef uint8 bool;\n'
enums.emit()

print '\n\n'

structs_used = entities.emit(fields, structs, enums)

# emit structs that weren't associated with an entity
structs.emit_unused(structs_used, fields, enums)




# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import utils

# eDB2FragmentType
TYPE_FIELD              = 0 # Simple field fragment
TYPE_STRUCT             = 1 # Structure fragment
TYPE_CONSTANT_PAD       = 2 # Pad fragment, fixed length (bits)
TYPE_VARIABLE_PAD_BITS  = 3 # Pad fragment, variable (bits)
TYPE_VARIABLE_PAD_BYTES = 4 # Pad fragment, variable (bytes)
TYPE_FULL_BYTE_PAD      = 5 # Pad fragment, pad to a full byte
TYPE_MSB_2_LSB          = 6 # Switch to MSB -> LSB order
TYPE_LSB_2_MSB          = 7 # Switch to LSB -> MSB order

def is_pad_type(t):
    if t == TYPE_CONSTANT_PAD or t == TYPE_VARIABLE_PAD_BITS or \
        t == TYPE_VARIABLE_PAD_BYTES or t == TYPE_FULL_BYTE_PAD:
        return True
    return False

# eDB2ModifierType
MOD_NONE             = 0  # Modifier is not used
MOD_CONSTANT_ARRAY   = 1  # Constant (elements) array
MOD_VARIABLE_ARRAY   = 2  # Variable (elements) array (modval gives #elements field)
MOD_OBSOLETE_3       = 3  # Constant (bits) array [OBS]
MOD_OBSOLETE_4       = 4  # Variable (bits) array [OBS]
MOD_OPTIONAL         = 5  # Fragment is optional
MOD_VARIABLE_ARRAY2  = 6  # Variable (elements) array, start/stop given
MOD_VARIABLE_ARRAY3  = 7  # Variable (elements) array, simple expression
MOD_VARIABLE_STRING1 = 8  # Variable length string (bit length)
MOD_VARIABLE_STRING2 = 9  # Variable length string (byte length)
MOD_VARIABLE_STRING3 = 10 # Variable length string (character length)


class FragmentBase:
    # Struct fragments (ie, each line in Struct.txt describe each member of
    # a struct.  The format is as follows:
    #
    # id^order^type^fieldid^name^offset^modtype^modval
    #
    # 50409^1^0^54024^""^-1^2^"54023"
    #
    # This describes the second element (ie, element #1) of struct 50409,
    # which is of type FIELD (0) and the value is described by Field #54024.
    # Additionally, it's "modtype" is VARIABLE_ARRAY which is described by
    # Field #54023.  Basically, this fragment is a variable array of
    # "Medium Preference" values, each element of which is one of the
    # "QMI PDS Mediums" (Enum.txt, #50407 given by Field.txt #54024).  The
    # number of elements in the array is given by the "modtype" and "modval"
    # pointers, which say that Field #54023 (which is also the first member
    # of this struct) specifies the number of elements in this variable array.

    def __init__(self, line):
        parts = line.split('^')
        if len(parts) < 8:
            raise Exception("Invalid struct fragment '%s'" % line)
        self.id = int(parts[0])
        self.order = int(parts[1])
        self.type = int(parts[2])   # eDB2FragmentType
        self.value = int(parts[3])  # id of field in Fields.txt
        self.name = parts[4].replace('"', '')
        self.offset = int(parts[5])
        self.modtype = int(parts[6])  # eDB2ModifierType
        self.modval = parts[7].replace('"', '')

    #Struct.txt:50021^0^0^50118^""^-1^1^"4"

    def validate(self, fields, structs):
        raise Exception("Surprising struct %d:%d field: %d" % (self.id, self.order, self.type))

    # Should return size in *bits* of this struct fragment, including all
    # sub-fragments
    def emit(self, do_print, entity_name, indent, reserved_bits, fields, structs, enums, cur_structsize):
        return 0

class Msb2LsbFragment(FragmentBase):
    # Subclass for TYPE_MSB_2_LSB
    def validate(self, fields, structs):
        pass

class FieldFragment(FragmentBase):
    # Subclass for TYPE_FIELD
    def validate(self, fields, structs):
        self.field = fields.get_child(self.value)

    def emit(self, do_print, entity_name, indent, reserved_bits, fields, structs, enums, cur_structsize):
        # modify the field like cProtocolEntityNav::ProcessFragment()
        num_elements = 0
        comment = ""
        isarray = False

        if self.modtype == MOD_CONSTANT_ARRAY:
            # modval is number of elements
            num_elements = int(self.modval)
        elif self.modtype == MOD_VARIABLE_ARRAY or \
             self.modtype == MOD_VARIABLE_STRING1 or \
             self.modtype == MOD_VARIABLE_STRING2 or \
             self.modtype == MOD_VARIABLE_STRING3:
            # The "modval" is the field id that gives the # of elements
            # of this variable array.  That field is usually the immediately
            # previous fragment of this struct.  ie it's something like:
            #
            # struct foobar {
            #     u8 arraylen;   <- specified by modval
            #     char array[];  <- current struct fragment
            # }
            #
            fdesc = fields.get_child(int(self.modval))
            comment = "size given by %s" % utils.nicename(fdesc.name)
            isarray = True

        return self.field.emit(do_print, indent, enums, num_elements, comment, isarray)


class StructFragment(FragmentBase):
    # Subclass for TYPE_STRUCT
    def validate(self, fields, structs):
        self.struct = structs.get_child(self.value)

    def emit(self, do_print, entity_name, indent, reserved_bits, fields, structs, enums, cur_structsize):
        # embedded structs often won't have a name of their own
        structname = "a_item"
        if self.name:
            structname = "%s_item" % utils.nicename(self.name)

        bits = self.struct.emit(structname, indent, reserved_bits, fields, structs, enums)

        # Ignore the condition on some structs' modvals
        fdesc_id = 0
        if len(self.modval) and self.modval.find("=") < 0:
            fdesc_id = int(self.modval)

        comment = ""
        arraybits = ""
        try:
            fdesc = fields.get_child(fdesc_id)
            comment = "\t /* size given by %s */" % utils.nicename(fdesc.name)
            arraybits = "[0]"
        except KeyError:
            pass

        if do_print:
            varname = utils.nicename(self.name)
            if not varname:
                varname = "item"
            print "%sstruct %s %s%s;%s" % ("\t" * indent, structname, utils.nicename(self.name), arraybits, comment)

        return bits

class ConstantPadFragment(FragmentBase):
    # Subclass for TYPE_CONSTANT_PAD
    def __init__(self, line):
        FragmentBase.__init__(self, line)
        # padsize is total struct size (in bits) including this fragment; ie
        # given the current struct size, pad the struct out to padsize
        self.padsize = self.value

    def validate(self, fields, structs):
        if self.value < 0 or self.value > 1000:
            raise Exception("Invalid constant pad size %d" % self.value)

    def emit(self, do_print, entity_name, indent, reserved_bits, fields, structs, enums, cur_structsize):
        # cur_structsize is in bits
        if cur_structsize > self.padsize:
            raise ValueError("Current structure size (%d) is larger than pad size (%d)!")
 
        padbits = self.padsize - cur_structsize
        padtype = "uint8"
        if padbits > 8 and padbits <= 16:
            padtype = "uint16"
        elif padbits > 16 and padbits <= 32:
            padtype = "uint32"
        elif padbits > 32 and padbits <= 64:
            padtype = "uint64"
        else:
            raise ValueError("FIXME: handle multi-long padding")

        if do_print:
            if padbits in [8, 16, 32]:
                print "%s%s padding;" % ("\t" * indent, padtype)
            else:
                print "%s%s padding:%d;" % ("\t" * indent, padtype, padbits)
        return padbits

class Struct:
    def __init__(self, sid):
        self.id = sid
        self.fragments = []
        self.name = "struct_%d" % sid

    def add_fragment(self, fragment):
        for f in self.fragments:
            if fragment.order == f.order:
                raise Exception("Struct %d already has fragment order %d" % (self.id, fragment.order))
        self.fragments.append(fragment)
        self.fragments.sort(lambda x, y: cmp(x.order, y.order))

    def validate(self, fields, structs):
        for f in self.fragments:
            f.validate(fields, structs)

    def emit_header(self, entity_name, cmdno, tlvno):
        if not entity_name:
            entity_name = self.name
            svcname = "Unknown"
            cmdname = "Unknown"
            tlvname = "Unknown"
        else:
            parts = entity_name.split("/")
            svcname = parts[0].upper()
            cmdname = parts[1].upper().replace(" ", "_")
            tlvname = parts[2]

        print '/**'
        print ' * SVC: %s' % svcname
        print ' * CMD: 0x%04x (%s)' % (cmdno, cmdname)
        print ' * TLV: 0x%02x   (%s)' % (tlvno, tlvname)
        print ' * ID:  %d' % self.id
        print ' */'

    def need_union(self, frag_idx):
        # scan to see if there is any pad fragment in this struct; if there is,
        # and if there's a struct fragment before the pad fragment, we'll need
        # to make a union
        found_struct = False
        for i in range(frag_idx, len(self.fragments)):
            if self.fragments[i].type == TYPE_STRUCT:
                found_struct = True
            if is_pad_type(self.fragments[i].type):
                if found_struct:
                    return True
                break
        return False

    def emit(self, entity_name, indent, pad_bits, fields, structs, enums):
        print '%sstruct %s {' % ("\t" * indent, utils.nicename(entity_name))

        # current structure size in *bits*
        size_in_bits = 0

        # if there was a previous member we're supposed to add some padding
        # for (ie, this struct is in a union and something is before it) do
        # that now
        if pad_bits:
            print '%suint8 padding:%d;' % ("\t" * (indent + 1), pad_bits)
            size_in_bits += pad_bits

        union_frag = None
        reserved_bits = 0

        for i in range(0, len(self.fragments)):
            frag = self.fragments[i]

            if not union_frag:
                if self.need_union(i):
                    union_frag = frag
                    indent += 1
                    print '%sunion u_%s {' % ("\t" * indent, utils.nicename(frag.field.name))

            bits = frag.emit(True, entity_name, indent + 1, reserved_bits, fields, structs, enums, size_in_bits)
            if frag == union_frag:
                # Track the first union fragment's reserved bits; we'll ignore
                # them for total struct size since the following struct(s) will
                # include it's size
                reserved_bits = bits
            elif is_pad_type(frag.type):
                # When we reach the pad fragment, the union terminates so
                # clear out union-specific state
                reserved_bits = 0
                if union_frag:
                    print '%s};' % ("\t" * indent)
                    indent -= 1
                    union_frag = None
                size_in_bits += bits
            else:
                size_in_bits += bits

        print "%s};\n" % ("\t" * indent)
        return size_in_bits

class Structs(utils.DbFile):
    def __init__(self, path):
        self.structs = {}
        f = file(path + "Struct.txt")
        for line in f:
            if len(line.strip()) == 0:
                continue

            # parse enough of the line to get struct ID and type
            parts = line.split('^')
            if len(parts) < 3:
                raise Exception("Invalid struct fragment '%s'" % line)

            # make a new struct if we don't know about this one already
            struct_id = int(parts[0])
            try:
                struct = self.structs[struct_id]
            except KeyError:
                struct = Struct(struct_id)
                self.structs[struct.id] = struct

            frag_type = int(parts[2])
            try:
                frag_class = { TYPE_FIELD: FieldFragment,
                               TYPE_STRUCT: StructFragment,
                               TYPE_CONSTANT_PAD: ConstantPadFragment,
                               TYPE_MSB_2_LSB: Msb2LsbFragment
                             }[frag_type]
            except KeyError:
                # fall back to base class
                frag_class = FragmentBase

            frag = frag_class(line.strip())
            struct.add_fragment(frag)
        f.close()

    def validate(self, fields):
        for s in self.structs.values():
            s.validate(fields, self)

    def has_child(self, sid):
        return sid in self.structs

    def get_child(self, sid):
        return self.structs[sid]

    def emit_unused(self, used, fields, enums):
        print '/**** UNKNOWN TLVs ****/\n'
        for s in self.structs.values():
            if not s.id in used:
                s.emit_header(None, 0, 0)
                s.emit(None, 0, fields, self, enums)




# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import utils

class EnumEntry:
    def __init__(self, line):
        parts = line.split('^')
        if len(parts) < 3:
            raise Exception("Invalid enum entry line '%s'" % line)
        self.id = int(parts[0])
        self.value = int(parts[1], 0)
        self.name = parts[2].replace('"', '')
        self.descid = None
        if len(parts) > 3:
            self.descid = int(parts[3])

    def emit(self, enum_name):
        print "\tGOBI_%s_%s\t\t= 0x%08x,     /* %s */" % (
                    utils.constname(enum_name),
                    utils.constname(self.name),
                    self.value, self.name)

class Enum:
    def __init__(self, line):
        parts = line.split('^')
        if len(parts) < 4:
            raise Exception("Invalid enum line '%s'" % line)
        self.id = int(parts[0])
        self.name = parts[1].replace('"', '')
        self.descid = int(parts[2])
        self.internal = int(parts[3]) != 0
        self.values = []   # list of EnumEntry objects

    def add_entry(self, entry):
        for v in self.values:
            if entry.value == v.value:
                raise Exception("Enum %d already has value %d" % (self.id, v.value))
        self.values.append(entry)
        self.values.sort(lambda x, y: cmp(x.value, y.value))

    def emit(self):
        print 'typedef enum { /* %s */ ' % self.name
        for en in self.values:
            en.emit(self.name)
        print "} gobi_%s;\n" % utils.nicename(self.name)
            

class Enums(utils.DbFile):
    def __init__(self, path):
        self.enums = {}

        # parse the enums
        f = file(path + "Enum.txt")
        for line in f:
            try:
                enum = Enum(line.strip())
                self.enums[enum.id] = enum
            except Exception(e):
                pass
        f.close()

        # and now the enum entries
        f = file(path + "EnumEntry.txt")
        for line in f:
            try:
                entry = EnumEntry(line.strip())
                self.enums[entry.id].add_entry(entry)
            except Exception(e):
                pass
        f.close()

    def emit(self):
        for e in self.enums:
            self.enums[e].emit()

    def get_child(self, eid):
        return self.enums[eid]




# -*- Mode: python; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2012 Lanedo GmbH
# Copyright (C) 2012-2022 Aleksander Morgado <aleksander@aleksander.es>
# Copyright (c) 2022 Qualcomm Innovation Center, Inc.
#

import string
import utils
from Variable import Variable

"""
Variable type for signed/unsigned Integers and floating point numbers:
 'guint8', 'gint8'
 'guint16', 'gint16'
 'guint32', 'gint32'
 'guint64', 'gint64'
 'guint-sized'
 'gfloat', 'gdouble'
"""
class VariableInteger(Variable):

    def __init__(self, service, dictionary):

        # Call the parent constructor
        Variable.__init__(self, service, dictionary)

        self.guint_sized_size = ''
        if self.format == "guint-sized":
            if 'guint-size' not in dictionary:
                raise RuntimeError('Format \'guint-sized\' requires \'guint-size\' parameter')
            else:
                self.guint_sized_size = dictionary['guint-size']
            self.private_format = 'guint64'
            self.public_format  = 'guint64'
        else:
            self.private_format = self.format
            self.public_format = dictionary['public-format'] if 'public-format' in dictionary else self.private_format
        self.element_type = self.public_format

        # Public format for the GIR array compat methods in 1.32
        self.private_format_gir = self.private_format
        self.public_format_gir = self.public_format
        self.element_type_gir = self.element_type


    def emit_buffer_read(self, f, line_prefix, tlv_out, error, variable_name):
        translations = { 'lp'             : line_prefix,
                         'tlv_out'        : tlv_out,
                         'variable_name'  : variable_name,
                         'error'          : error,
                         'public_format'  : self.public_format,
                         'private_format' : self.private_format,
                         'len'            : self.guint_sized_size }

        if self.private_format not in ('guint8', 'gint8'):
            translations['endian'] = ' ' + self.endian + ','
        else:
            translations['endian'] = ''

        if self.format == 'guint-sized':
            template = (
                '${lp}if (!qmi_message_tlv_read_sized_guint (message, init_offset, &offset, ${len},${endian} &(${variable_name}), ${error}))\n'
                '${lp}    goto ${tlv_out};\n')
        elif self.format == 'gfloat':
            template = (
                '${lp}if (!qmi_message_tlv_read_gfloat_endian (message, init_offset, &offset,${endian} &(${variable_name}), ${error}))\n'
                '${lp}    goto ${tlv_out};\n')
        elif self.private_format == self.public_format:
            template = (
                '${lp}if (!qmi_message_tlv_read_${private_format} (message, init_offset, &offset,${endian} &(${variable_name}), ${error}))\n'
                '${lp}    goto ${tlv_out};\n')
        else:
            template = (
                '${lp}{\n'
                '${lp}    ${private_format} tmp;\n'
                '\n'
                '${lp}    if (!qmi_message_tlv_read_${private_format} (message, init_offset, &offset,${endian} &tmp, ${error}))\n'
                '${lp}        goto ${tlv_out};\n'
                '${lp}    ${variable_name} = (${public_format})tmp;\n'
                '${lp}}\n')
        f.write(string.Template(template).substitute(translations))


    @staticmethod
    def fixed_type_byte_size(fmt):
        if fmt == 'guint8':
            return 1
        if fmt == 'guint16':
            return 2
        if fmt == 'guint32':
            return 4
        if fmt == 'guint64':
            return 8
        if fmt == 'gint8':
            return 1
        if fmt == 'gint16':
            return 2
        if fmt == 'gint32':
            return 4
        if fmt == 'gint64':
            return 8
        raise Exception("Unsupported format %s" % (fmt))


    def emit_buffer_write(self, f, line_prefix, tlv_name, variable_name):
        translations = { 'lp'             : line_prefix,
                         'private_format' : self.private_format,
                         'len'            : self.guint_sized_size,
                         'tlv_name'       : tlv_name,
                         'variable_name'  : variable_name }

        if self.private_format != 'guint8' and self.private_format != 'gint8':
            translations['endian'] = ' ' + self.endian + ','
        else:
            translations['endian'] = ''

        if self.format == 'guint-sized':
            template = (
                '${lp}/* Write the ${len}-byte long variable to the buffer */\n'
                '${lp}if (!qmi_message_tlv_write_sized_guint (self, ${len},${endian} ${variable_name}, error)) {\n'
                '${lp}    g_prefix_error (error, "Cannot write sized integer in TLV \'${tlv_name}\': ");\n'
                '${lp}    return NULL;\n'
                '${lp}}\n')
        elif self.private_format == self.public_format:
            template = (
                '${lp}/* Write the ${private_format} variable to the buffer */\n'
                '${lp}if (!qmi_message_tlv_write_${private_format} (self,${endian} ${variable_name}, error)) {\n'
                '${lp}    g_prefix_error (error, "Cannot write integer in TLV \'${tlv_name}\': ");\n'
                '${lp}    return NULL;\n'
                '${lp}}\n')
        else:
            template = (
                '${lp}{\n'
                '${lp}    ${private_format} tmp;\n'
                '\n'
                '${lp}    tmp = (${private_format}) ${variable_name};\n'
                '${lp}    /* Write the ${private_format} variable to the buffer */\n'
                '${lp}    if (!qmi_message_tlv_write_${private_format} (self,${endian} tmp, error)) {\n'
                '${lp}        g_prefix_error (error, "Cannot write enum in TLV \'${tlv_name}\': ");\n'
                '${lp}        return NULL;\n'
                '${lp}    }\n'
                '${lp}}\n')
        f.write(string.Template(template).substitute(translations))


    def emit_get_printable(self, f, line_prefix, is_personal):
        common_format = ''
        common_cast = ''

        if self.private_format == 'guint8':
            common_format = '%u'
            common_cast = '(guint)'
        elif self.private_format == 'guint16':
            common_format = '%" G_GUINT16_FORMAT "'
        elif self.private_format == 'guint32':
            common_format = '%" G_GUINT32_FORMAT "'
        elif self.private_format == 'guint64':
            common_format = '%" G_GUINT64_FORMAT "'
        elif self.private_format == 'gint8':
            common_format = '%d'
            common_cast = '(gint)'
        elif self.private_format == 'gint16':
            common_format = '%" G_GINT16_FORMAT "'
        elif self.private_format == 'gint32':
            common_format = '%" G_GINT32_FORMAT "'
        elif self.private_format == 'gint64':
            common_format = '%" G_GINT64_FORMAT "'
        elif self.private_format in ('gfloat', 'gdouble'):
            common_format = '%lf'
            common_cast = '(gdouble)'

        translations = { 'lp'             : line_prefix,
                         'private_format' : self.private_format,
                         'public_format'  : self.public_format,
                         'len'            : self.guint_sized_size,
                         'common_format'  : common_format,
                         'common_cast'    : common_cast }

        if self.private_format not in ('guint8', 'gint8'):
            translations['endian'] = ' ' + self.endian + ','
        else:
            translations['endian'] = ''

        template = (
            '\n'
            '${lp}{\n'
            '${lp}    ${private_format} tmp;\n'
            '\n')

        if self.format == 'guint-sized':
            template += (
                '${lp}    if (!qmi_message_tlv_read_sized_guint (message, init_offset, &offset, ${len},${endian} &tmp, &error))\n'
                '${lp}        goto out;\n')
        elif self.format == 'gfloat':
            template += (
                '${lp}    if (!qmi_message_tlv_read_gfloat_endian (message, init_offset, &offset,${endian} &tmp, &error))\n'
                '${lp}        goto out;\n')
        else:
            template += (
                '${lp}    if (!qmi_message_tlv_read_${private_format} (message, init_offset, &offset,${endian} &tmp, &error))\n'
                '${lp}        goto out;\n')

        if self.personal_info or is_personal:
            translations['if_show_field'] = 'if (qmi_utils_get_show_personal_info ()) '
        else:
            translations['if_show_field'] = ''

        template += (
            '${lp}    ${if_show_field}{\n')
        if self.public_format == 'gboolean':
            template += (
                '${lp}        g_string_append_printf (printable, "%s", tmp ? "yes" : "no");\n')
        elif self.public_format != self.private_format:
            translations['public_type_underscore'] = utils.build_underscore_name_from_camelcase(self.public_format)
            translations['public_type_underscore_upper'] = utils.build_underscore_name_from_camelcase(self.public_format).upper()
            template += (
                '#if defined  __${public_type_underscore_upper}_IS_ENUM__\n'
                '${lp}        g_string_append_printf (printable, "%s", ${public_type_underscore}_get_string ((${public_format})tmp));\n'
                '#elif defined  __${public_type_underscore_upper}_IS_FLAGS__\n'
                '${lp}        {\n'
                '${lp}            g_autofree gchar *flags_str = NULL;\n'
                '\n'
                '${lp}            flags_str = ${public_type_underscore}_build_string_from_mask ((${public_format})tmp);\n'
                '${lp}            g_string_append_printf (printable, "%s", flags_str);\n'
                '${lp}        }\n'
                '#else\n'
                '# error unexpected public format: ${public_format}\n'
                '#endif\n')
        else:
            template += (
                '${lp}        g_string_append_printf (printable, "${common_format}", ${common_cast}tmp);\n')

        if self.personal_info or is_personal:
            template += (
                '${lp}    } else {\n'
                '${lp}        g_string_append_printf (printable, "\'###\'");\n')

        template += (
            '${lp}    }\n'
            '${lp}}\n')

        f.write(string.Template(template).substitute(translations))


    def build_variable_declaration(self, line_prefix, variable_name):
        translations = { 'lp'             : line_prefix,
                         'private_format' : self.private_format,
                         'name'           : variable_name }
        template = (
            '${lp}${private_format} ${name};\n')
        return string.Template(template).substitute(translations)


    def build_struct_field_declaration(self, line_prefix, variable_name):
        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }
        template = (
            '${lp}${public_format} ${name};\n')
        return string.Template(template).substitute(translations)


    def build_getter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }

        template = (
            '${lp}${public_format} *${name},\n')
        return string.Template(template).substitute(translations)


    def build_getter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }

        template = (
            '${lp}@${name}: (out)(optional): a placeholder for the output #${public_format}, or %NULL if not required.\n')
        return string.Template(template).substitute(translations)


    def build_getter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        needs_cast = True if self.public_format != self.private_format else False
        translations = { 'lp'       : line_prefix,
                         'from'     : variable_name_from,
                         'to'       : variable_name_to,
                         'cast_ini' : '(' + self.public_format + ')(' if needs_cast else '',
                         'cast_end' : ')' if needs_cast else '' }

        template = (
            '${lp}if (${to})\n'
            '${lp}    *${to} = ${cast_ini}${from}${cast_end};\n')
        return string.Template(template).substitute(translations)


    def build_setter_declaration(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }

        template = (
            '${lp}${public_format} ${name},\n')
        return string.Template(template).substitute(translations)


    def build_setter_documentation(self, line_prefix, variable_name):
        if not self.visible:
            return ""

        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }

        template = (
            '${lp}@${name}: a #${public_format}.\n')
        return string.Template(template).substitute(translations)


    def build_setter_implementation(self, line_prefix, variable_name_from, variable_name_to):
        if not self.visible:
            return ""

        needs_cast = True if self.public_format != self.private_format else False
        translations = { 'lp'       : line_prefix,
                         'from'     : variable_name_from,
                         'to'       : variable_name_to,
                         'cast_ini' : '(' + self.private_format + ')(' if needs_cast else '',
                         'cast_end' : ')' if needs_cast else '' }

        template = (
            '${lp}${to} = ${cast_ini}${from}${cast_end};\n')
        return string.Template(template).substitute(translations)


    def build_struct_field_documentation(self, line_prefix, variable_name):
        translations = { 'lp'            : line_prefix,
                         'public_format' : self.public_format,
                         'name'          : variable_name }

        template = (
            '${lp}@${name}: a #${public_format}.\n')
        return string.Template(template).substitute(translations)


    def build_copy_gir(self, line_prefix, variable_name_from, variable_name_to):
        translations = { 'lp'                 : line_prefix,
                         'variable_name_from' : variable_name_from,
                         'variable_name_to'   : variable_name_to }

        template = (
            '${lp}${variable_name_to} = ${variable_name_from};\n')
        return string.Template(template).substitute(translations)


    def build_copy_to_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_copy_gir (line_prefix, variable_name_from, variable_name_to)


    def build_copy_from_gir(self, line_prefix, variable_name_from, variable_name_to):
        return self.build_copy_gir (line_prefix, variable_name_from, variable_name_to)
